# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Iterable, List, Optional, Tuple

from spinn_utilities.overrides import overrides
from spinn_utilities.config_holder import set_config

from pacman.model.graphs.application import (
    ApplicationEdge, ApplicationEdgePartition, ApplicationFPGAVertex,
    FPGAConnection)
from pacman.model.graphs.machine import MachineVertex
from pacman.utilities.utility_calls import get_keys

from spinn_front_end_common.abstract_models import (
    AbstractSendMeMulticastCommandsVertex, LiveOutputDevice,
    HasCustomAtomKeyMap)
from spinn_front_end_common.utility_models.command_sender import CommandSender
from spinn_front_end_common.utility_models import MultiCastCommand

from spynnaker.pyNN.models.common import PopulationApplicationVertex
from spynnaker.pyNN.models.populations import Population
from spynnaker.pyNN.data.spynnaker_data_view import SpynnakerDataView
from spynnaker.pyNN.spynnaker_external_device_plugin_manager import (
    SpynnakerExternalDevicePluginManager)

from .spif_devices import (
    SPIF_FPGA_ID, SPIF_OUTPUT_FPGA_LINK,
    set_distiller_key, set_distiller_mask,
    set_distiller_mask_delayed, set_distiller_shift,
    set_xp_key_delayed, set_xp_mask_delayed)

# The maximum number of partitions that can be supported.
N_OUTGOING = 6


class SPIFOutputDevice(
        ApplicationFPGAVertex, PopulationApplicationVertex,
        AbstractSendMeMulticastCommandsVertex, LiveOutputDevice):
    """
    Output (only) to a SPIF device.  Each SPIF device can accept up to 6
    incoming projections.
    Keys sent from Populations to SPIF will be mapped by removing the
    SpiNNaker key and adding an index so that the source Population can
    be identified.  Source Populations must be split into power-of-two
    sized cores to ensure that keys are contiguous.
    The keys output by SPIF will be of the form:

         (projection_index << output_key_shift) | neuron_id

    By default, the projection index will be in the top 8 bits of the
    packet, but this can be controlled with the output_key_shift parameter.
    """

    __slots__ = ("__incoming_partitions", "__create_database",
                 "__output_key_shift", "__output_key_and_mask")

    def __init__(self, board_address: Optional[str] = None,
                 chip_coords: Optional[Tuple[int, int]] = None,
                 label: Optional[str] = None,
                 create_database: bool = True,
                 database_notify_host: Optional[str] = None,
                 database_notify_port_num: Optional[int] = None,
                 database_ack_port_num: Optional[int] = None,
                 output_key_shift: int = 24):
        """
        :param board_address: The board IP address of the SPIF device
        :param chip_coords: The chip coordinates of the SPIF device
        :param label: The label to give the SPIF device
        :param create_database:
            Whether the database will be used to decode keys or not
        :param database_notify_host: The host that will read the database
        :param database_notify_port_num:
            The port of the host that will read the database
        :param database_ack_port_num:
            The port to listen on for responses from the host reading the
            database
        :param output_key_shift:
            The shift to apply to the population indices when added to the key
        """
        super(SPIFOutputDevice, self).__init__(
            n_atoms=1,
            outgoing_fpga_connection=FPGAConnection(
                SPIF_FPGA_ID, SPIF_OUTPUT_FPGA_LINK, board_address,
                chip_coords),
            label=label)
        self.__incoming_partitions: List[ApplicationEdgePartition] = list()
        # Force creation of the database, to be used in the read side of things
        if create_database:
            set_config("Database", "create_database", "True")
            SpynnakerExternalDevicePluginManager.add_database_socket_address(
                database_notify_host, database_notify_port_num,
                database_ack_port_num)
        self.__create_database = create_database
        self.__output_key_shift = output_key_shift
        self.__output_key_and_mask: \
            Dict[PopulationApplicationVertex, Tuple[int, int]] = dict()

    def set_output_key_and_mask(
            self, population: Population, key: int, mask: int) -> None:
        """ Set the output key to be written into packets when received by
            SPIF, and the mask to apply before adding the key.  The key should
            be the exact value that will be "or'ed" with the packet after
            masking.  The mask should be the mask to apply to the incoming
            SpiNNaker key to extract the neuron id bits.  The key and mask
            will not be checked; please make sure you are using values that
            make sense!

        :param population: The PyNN source Population
        :param key: The key to "or" with the incoming key *after* masking
        :param mask: The mask to "and" with the incoming SpiNNaker key
        """
        # pylint: disable=protected-access
        self.__output_key_and_mask[population._vertex] = (key, mask)

    def __is_power_of_2(self, v: int) -> bool:
        """ Determine if a value is a power of 2.

        :param v: The value to test
        """
        return (v & (v - 1) == 0) and (v != 0)

    @overrides(ApplicationFPGAVertex.add_incoming_edge)
    def add_incoming_edge(self, edge: ApplicationEdge,
                          partition: ApplicationEdgePartition) -> None:
        # Only add edges from PopulationApplicationVertices
        if not isinstance(edge.pre_vertex, PopulationApplicationVertex):
            if not isinstance(edge.pre_vertex, CommandSender):
                raise ValueError(
                    "This vertex only accepts input from "
                    "PopulationApplicationVertex instances")
            # Ignore the command sender sending to us!
            return

        if len(self.__incoming_partitions) >= N_OUTGOING:
            raise ValueError(
                f"Only {N_OUTGOING} outgoing connections are supported per"
                " spif device (existing partitions:"
                f" {self.__incoming_partitions}")
        # Ensure the incoming thing is split appropriately, as otherwise keys
        # won't be correct
        max_atoms = partition.pre_vertex.get_max_atoms_per_core()
        if max_atoms < partition.pre_vertex.n_atoms:
            if not self.__is_power_of_2(max_atoms):
                raise ValueError(
                    "The incoming vertex will be split into units of"
                    f" {max_atoms}, which means that the keys won't be"
                    " contiguous.  Please choose a power-of-two size for the"
                    " maximum atoms per core")
        self.__incoming_partitions.append(partition)
        if self.__create_database and len(self.__incoming_partitions) == 1:
            SpynnakerDataView.add_live_output_device(self)

    def _get_set_key_payload(self, index: int) -> int:
        """
        Get the payload for the command to set the router key.

        :param index: The index of key to get
        """
        r_infos = SpynnakerDataView.get_routing_infos()
        return r_infos.get_key_from(
            self.__incoming_partitions[index].pre_vertex,
            self.__incoming_partitions[index].identifier)

    def _get_set_mask_payload(self, index: int) -> int:
        """
        Get the payload for the command to set the router mask.

        :param index: The index of the mask to get
        """
        r_infos = SpynnakerDataView.get_routing_infos()
        return r_infos.get_info_from(
            self.__incoming_partitions[index].pre_vertex,
            self.__incoming_partitions[index].identifier).mask

    def _get_set_dist_mask_payload(self, index: int) -> int:
        """ Get the payload for the command to set the distiller mask
        """
        r_infos = SpynnakerDataView.get_routing_infos()
        return ~r_infos.get_info_from(
            self.__incoming_partitions[index].pre_vertex,
            self.__incoming_partitions[index].identifier).mask & 0xFFFFFFFF

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.start_resume_commands)
    def start_resume_commands(self) -> Iterable[MultiCastCommand]:
        # The commands here are delayed, as at the time of providing them,
        # we don't know the key or mask of the incoming link...
        commands: List[MultiCastCommand] = list()
        for i, part in enumerate(self.__incoming_partitions):
            pop_vertex = part.pre_vertex
            assert isinstance(pop_vertex, PopulationApplicationVertex)
            commands.append(set_xp_key_delayed(i, self._get_set_key_payload))
            commands.append(set_xp_mask_delayed(i, self._get_set_mask_payload))
            if pop_vertex in self.__output_key_and_mask:
                key, mask = self.__output_key_and_mask[pop_vertex]
                commands.append(set_distiller_key(i, key))
                commands.append(set_distiller_mask(i, mask))
            else:
                commands.append(set_distiller_key(
                    i, i << self.__output_key_shift))
                commands.append(set_distiller_mask_delayed(
                    i, self._get_set_dist_mask_payload))
            commands.append(set_distiller_shift(
                i, pop_vertex.n_colour_bits))
        return commands

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.pause_stop_commands)
    def pause_stop_commands(self) -> Iterable[MultiCastCommand]:
        return []

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.timed_commands)
    def timed_commands(self) -> List[MultiCastCommand]:
        return []

    @overrides(LiveOutputDevice.get_device_output_keys)
    def get_device_output_keys(self) -> Dict[MachineVertex,
                                             List[Tuple[int, int]]]:
        all_keys: Dict[MachineVertex, List[Tuple[int, int]]] = dict()
        routing_infos = SpynnakerDataView.get_routing_infos()
        for i, part in enumerate(self.__incoming_partitions):
            pop_vertex = part.pre_vertex
            assert isinstance(pop_vertex, PopulationApplicationVertex)
            if pop_vertex in self.__output_key_and_mask:
                key, mask = self.__output_key_and_mask[pop_vertex]
            else:
                key = i << self.__output_key_shift
                mask = self._get_set_dist_mask_payload(i)
            shift = pop_vertex.n_colour_bits
            for m_vertex in part.pre_vertex.splitter.get_out_going_vertices(
                    part.identifier):
                atom_keys: Iterable[Tuple[int, int]] = list()
                if isinstance(m_vertex.app_vertex, HasCustomAtomKeyMap):
                    atom_keys = m_vertex.app_vertex.get_atom_key_map(
                        m_vertex, part.identifier, routing_infos)
                else:
                    r_info = \
                        routing_infos.get_info_from(
                            m_vertex, part.identifier)
                    vertex_slice = m_vertex.vertex_slice
                    keys = get_keys(r_info.key, vertex_slice)
                    start = vertex_slice.lo_atom
                    atom_keys = [(i, k) for i, k in enumerate(keys, start)]

                atom_keys_mapped = list((i, key | ((k & mask) >> shift))
                                        for i, k in atom_keys)
                all_keys[m_vertex] = atom_keys_mapped
        return all_keys



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from numpy import floating
from numpy.typing import NDArray
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_front_end_common.interface.ds import DataSpecificationBase
from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    AbstractHasParameterNames)
from spynnaker.pyNN.models.neuron.plasticity.stdp.synapse_structure import (
    AbstractSynapseStructure)


class AbstractTimingDependence(
        AbstractHasParameterNames, metaclass=AbstractBase):
    """
    An STDP timing dependence rule.
    """
    __slots__ = ("__synapse_structure", )

    def __init__(self, synapse_structure: AbstractSynapseStructure):
        """
        :param synapse_structure:
            The synaptic structure of the plastic part of the rows.
        """
        self.__synapse_structure = synapse_structure

    @abstractmethod
    def is_same_as(
            self, timing_dependence: 'AbstractTimingDependence') -> bool:
        """
        :param timing_dependence:
        :returns: True if the timing dependence is the same as another.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def vertex_executable_suffix(self) -> str:
        """
        The suffix to be appended to the vertex executable for this rule.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def pre_trace_n_bytes(self) -> int:
        """
        The number of bytes used by the pre-trace of the rule per neuron.
        """
        raise NotImplementedError

    @abstractmethod
    def get_parameters_sdram_usage_in_bytes(self) -> int:
        """
        :returns: The amount of SDRAM used by the parameters of this rule.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def n_weight_terms(self) -> int:
        """
        The number of weight terms expected by this timing rule.
        """
        raise NotImplementedError

    @abstractmethod
    def write_parameters(
            self, spec: DataSpecificationBase, global_weight_scale: float,
            synapse_weight_scales: NDArray[floating]) -> None:
        """
        Write the parameters of the rule to the spec.

        :param spec: The specification to write to
        :param global_weight_scale: The weight scale applied globally
        :param synapse_weight_scales:
            The total weight scale applied to each synapse including the global
            weight scale
        """
        raise NotImplementedError

    @property
    def synaptic_structure(self) -> AbstractSynapseStructure:
        """
        The synaptic structure of the plastic part of the rows.
        """
        return self.__synapse_structure

    @property
    @abstractmethod
    def A_plus(self) -> float:
        r"""
        :math:`A^+`
        """
        # pylint: disable=invalid-name
        raise NotImplementedError

    @property
    @abstractmethod
    def A_minus(self) -> float:
        r"""
        :math:`A^-`
        """
        # pylint: disable=invalid-name
        raise NotImplementedError



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import List, Sequence, TYPE_CHECKING, Union, Optional
from spinn_utilities.config_holder import get_config_bool
from pacman.model.graphs.application import ApplicationVertex
from spynnaker.pyNN.models.neural_projections.connectors import (
    AbstractConnector, AbstractGenerateConnectorOnMachine)
from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    AbstractGenerateOnMachine)
from spynnaker.pyNN.types import (DELAYS, WEIGHTS)
from spynnaker.pyNN.utilities.constants import SPIKE_PARTITION_ID
if TYPE_CHECKING:
    from spynnaker.pyNN.models.populations import Population, PopulationView
    from spynnaker.pyNN.models.neuron import ConnectionHolder
    from spynnaker.pyNN.models.neuron.synapse_dynamics import (
        AbstractSynapseDynamics)


class SynapseInformation(object):
    """
    Contains the synapse information including the connector, synapse type
    and synapse dynamics.
    """
    # Made by a Projection
    __slots__ = (
        "__connector",
        "__pre_population",
        "__post_population",
        "__prepop_is_view",
        "__postpop_is_view",
        "__synapse_dynamics",
        "__synapse_type",
        "__receptor_type",
        "__weights",
        "__delays",
        "__pre_run_connection_holders",
        "__synapse_type_from_dynamics",
        "__download_on_pause",
        "__partition_id")

    def __init__(self, connector: AbstractConnector,
                 pre_population: Union[Population, PopulationView],
                 post_population: Union[Population, PopulationView],
                 prepop_is_view: bool, postpop_is_view: bool,
                 synapse_dynamics: AbstractSynapseDynamics,
                 synapse_type: int, receptor_type: str,
                 synapse_type_from_dynamics: bool,
                 weights: WEIGHTS = None,
                 delays: DELAYS = None,
                 download_on_pause: bool = False,
                 partition_id: Optional[str] = None):
        """
        :param connector: The connector connected to the synapse
        :param pre_population: The population sending spikes to the synapse
        :param post_population: The population hosting the synapse
        :param prepop_is_view: Whether the ``pre_population`` is a view
        :param postpop_is_view: Whether the ``post_population`` is a view
        :param synapse_dynamics: The dynamic behaviour of the synapse
        :param synapse_type: The type of the synapse
        :param receptor_type: Description of the receptor (e.g. excitatory)
        :param synapse_type_from_dynamics:
            Whether the synapse type came from synapse dynamics
        :param weights: The synaptic weights
        :param delays: The total synaptic delays
        :param download_on_pause:
            Whether to download the synapse matrix when the simulation pauses
        :param partition_id:
            The partition id for the application edge when not standard; if
            None, the standard SPIKE_PARTITION_ID is used
        """
        self.__connector = connector
        self.__pre_population = pre_population
        self.__post_population = post_population
        self.__prepop_is_view = prepop_is_view
        self.__postpop_is_view = postpop_is_view
        self.__synapse_dynamics = synapse_dynamics
        self.__synapse_type = synapse_type
        self.__receptor_type = receptor_type
        assert (delays is not None)
        self.__weights = weights
        self.__delays = delays
        self.__synapse_type_from_dynamics = synapse_type_from_dynamics
        self.__download_on_pause = download_on_pause
        self.__partition_id = partition_id or SPIKE_PARTITION_ID

        # Make a list of holders to be updated
        self.__pre_run_connection_holders: List[ConnectionHolder] = list()

    @property
    def connector(self) -> AbstractConnector:
        """
        The connector connected to the synapse.
        """
        return self.__connector

    @property
    def pre_population(self) -> Union[Population, PopulationView]:
        """
        The population sending spikes to the synapse.
        """
        return self.__pre_population

    @property
    def post_population(self) -> Union[Population, PopulationView]:
        """
        The population hosting the synapse.
        """
        return self.__post_population

    @property
    def pre_vertex(self) -> ApplicationVertex:
        """
        The vertex sending spikes to the synapse.
        """
        # pylint: disable=protected-access
        return self.__pre_population._vertex

    @property
    def post_vertex(self) -> ApplicationVertex:
        """
        The vertex hosting the synapse.
        """
        # pylint: disable=protected-access
        return self.__post_population._vertex

    @property
    def n_pre_neurons(self) -> int:
        """
        The number of neurons in the pre-population.
        """
        return self.__pre_population.size

    @property
    def n_post_neurons(self) -> int:
        """
        The number of neurons in the post-population.
        """
        return self.__post_population.size

    @property
    def prepop_is_view(self) -> bool:
        """
        Whether the :py:meth:`pre_population` is a view.
        """
        return self.__prepop_is_view

    @property
    def postpop_is_view(self) -> bool:
        """
        Whether the :py:meth:`post_population` is a view.
        """
        return self.__postpop_is_view

    @property
    def synapse_dynamics(self) -> AbstractSynapseDynamics:
        """
        The dynamic behaviour of the synapse.
        """
        return self.__synapse_dynamics

    @property
    def synapse_type(self) -> int:
        """
        The type of the synapse. An index into the set of synapse types
        supported by a neuron.
        """
        return self.__synapse_type

    @property
    def receptor_type(self) -> str:
        """
        A string representing the receptor type.
        """
        return self.__receptor_type

    @property
    def weights(self) -> WEIGHTS:
        """
        The synaptic weights (if any).
        """
        return self.__weights

    @property
    def delays(self) -> DELAYS:
        """
        The total synaptic delays (if any).
        """
        return self.__delays

    def may_generate_on_machine(self) -> bool:
        """
        Do we describe a collection of synapses whose synaptic matrix may
        be generated on SpiNNaker instead of needing to be calculated in
        this process and uploaded? This depends on the connector, the
        definitions of the weights and delays, and the dynamics of the
        synapses.

        :return: True if the synaptic matrix may be generated on machine (or
            may have already been so done)
        """
        # If we are using a virtual machine, we can't generate on the machine
        if get_config_bool("Machine", "virtual_board"):
            return False
        connector_gen = (
            isinstance(self.connector, AbstractGenerateConnectorOnMachine) and
            self.connector.generate_on_machine(self))
        synapse_gen = (
            isinstance(self.synapse_dynamics, AbstractGenerateOnMachine) and
            self.synapse_dynamics.generate_on_machine())
        return connector_gen and synapse_gen

    @property
    def pre_run_connection_holders(self) -> Sequence[ConnectionHolder]:
        """
        The list of connection holders to be filled in before run.
        """
        return self.__pre_run_connection_holders

    def add_pre_run_connection_holder(
            self, pre_run_connection_holder: ConnectionHolder) -> None:
        """
        Add a connection holder that will be filled in before run.

        :param pre_run_connection_holder:
            The connection holder to be added
        """
        self.__pre_run_connection_holders.append(pre_run_connection_holder)

    def finish_connection_holders(self) -> None:
        """
        Finish all the connection holders, and clear the list so that they
        are not generated again later.
        """
        for holder in self.__pre_run_connection_holders:
            holder.finish()
        del self.__pre_run_connection_holders[:]

    @property
    def synapse_type_from_dynamics(self) -> bool:
        """
        Whether the synapse type comes from the synapse dynamics.
        """
        return self.__synapse_type_from_dynamics

    @property
    def download_on_pause(self) -> bool:
        """
        Whether to download the synapse matrix when the simulation pauses.
        """
        return self.__download_on_pause

    @download_on_pause.setter
    def download_on_pause(self, download_on_pause: bool) -> None:
        """
        Set whether to download the synapse matrix when the simulation pauses.

        :param download_on_pause:
            Whether to download the synapse matrix when the simulation pauses
        """
        self.__download_on_pause = download_on_pause

    @property
    def partition_id(self) -> str:
        """
        The partition id for the application edge
        """
        return self.__partition_id



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A population is a group of neurons with the same neuron model and synaptic
model, but possibly (usually!) varying connectivity and configuration
parameters.

A population view is a subset of a population, created by slicing the
population::

    view = population[n:m]

An assembly is an agglomeration of populations and population views, created
by adding them together::

    assembly = population_1 + population_2

.. note::
    sPyNNaker only has incomplete support for assemblies; do not use.
"""

from .assembly import Assembly
from .population_base import PopulationBase
from .population import Population
from .population_view import PopulationView, IDMixin

__all__ = ["Assembly", "IDMixin", "Population", "PopulationBase",
           "PopulationView"]



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .abstract_connector import AbstractConnector
from .abstract_generate_connector_on_machine import (
    AbstractGenerateConnectorOnMachine)
from .abstract_generate_connector_on_host import (
    AbstractGenerateConnectorOnHost)
from .all_but_me_connector import AllButMeConnector
from .all_to_all_connector import AllToAllConnector
from .array_connector import ArrayConnector
from .csa_connector import CSAConnector
from .distance_dependent_probability_connector import (
    DistanceDependentProbabilityConnector)
from .fixed_number_post_connector import FixedNumberPostConnector
from .fixed_number_pre_connector import FixedNumberPreConnector
from .fixed_probability_connector import FixedProbabilityConnector
from .from_file_connector import FromFileConnector
from .from_list_connector import FromListConnector
from .index_based_probability_connector import IndexBasedProbabilityConnector
from .multapse_connector import MultapseConnector
from .one_to_one_connector import OneToOneConnector
from .small_world_connector import SmallWorldConnector
from .kernel_connector import KernelConnector
from .convolution_connector import ConvolutionConnector
from .pool_dense_connector import PoolDenseConnector
from .one_to_one_offset_connector import OneToOneOffsetConnector

__all__ = ["AbstractConnector", "AbstractGenerateConnectorOnMachine",
           "AbstractGenerateConnectorOnHost", "AllButMeConnector",
           "AllToAllConnector", "ArrayConnector", "CSAConnector",
           "DistanceDependentProbabilityConnector", "FixedNumberPostConnector",
           "FixedNumberPreConnector", "FixedProbabilityConnector",
           "FromFileConnector",
           "FromListConnector", "IndexBasedProbabilityConnector",
           "KernelConnector", "ConvolutionConnector", "PoolDenseConnector",
           "MultapseConnector", "OneToOneConnector", "SmallWorldConnector",
           "OneToOneOffsetConnector"]



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Sequence, Union

import numpy
from numpy.typing import NDArray

from spinn_utilities.classproperty import classproperty
from spinn_utilities.overrides import overrides

from pacman.model.partitioner_splitters import AbstractSplitterCommon

from spynnaker.pyNN.models.abstract_pynn_model import AbstractPyNNModel

from .spike_source_poisson_vertex import SpikeSourcePoissonVertex

_population_parameters = {"seed": None, "splitter": None}

DEFAULT_MAX_ATOMS_PER_CORE = 500


class SpikeSourcePoissonVariable(AbstractPyNNModel):
    """
    Model to create a Spike Source Poisson Vertex.
    """
    default_population_parameters = _population_parameters

    def __init__(
            self, rates: Union[Sequence[float], NDArray[numpy.floating], None],
            starts: Union[Sequence[int], NDArray[numpy.integer]],
            durations: Union[
                Sequence[int], NDArray[numpy.integer], None] = None):
        """
=
        :param rates: The spike rate of each neuron
        :param starts:  The start time of spikes on each neuron
        :param durations: The duration of spikes on each neuron
        """
        self._rates = rates
        self._starts = starts
        self._durations = durations

    @classproperty
    def absolute_max_atoms_per_core(  # pylint: disable=no-self-argument
            cls) -> int:
        return DEFAULT_MAX_ATOMS_PER_CORE

    @overrides(AbstractPyNNModel.create_vertex)
    def create_vertex(
            self, n_neurons: int, label: str, *, seed: Optional[int] = None,
            splitter: Optional[AbstractSplitterCommon] = None
            ) -> SpikeSourcePoissonVertex:
        """
        :param seed:
        :param splitter:
        """
        max_atoms = self.get_model_max_atoms_per_dimension_per_core()
        return SpikeSourcePoissonVertex(
            n_neurons, label, seed, max_atoms, self, rates=self._rates,
            starts=self._starts, durations=self._durations, splitter=splitter)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import logging
import math
import re
from typing import Dict, Optional, Sequence, Tuple, Union, TYPE_CHECKING
from typing_extensions import Never

import numpy
from numpy import float64, uint32, uint16, uint8
from numpy.typing import NDArray

from pyNN.random import NumpyRNG, RandomDistribution
from pyNN.space import Space

from spinn_utilities.log import FormatAdapter
from spinn_utilities.logger_utils import warn_once
from spinn_utilities.safe_eval import SafeEval
from spinn_utilities.abstract_base import AbstractBase, abstractmethod

from pacman.model.graphs import AbstractVertex
from pacman.model.graphs.common import Slice
from pacman.model.graphs.application import ApplicationVertex
from pacman.model.graphs.machine import MachineVertex

from spinn_front_end_common.interface.provenance import ProvenanceWriter
from spinn_front_end_common.utilities.exceptions import ConfigurationException

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.types import (
    DELAYS, is_scalar, WEIGHTS_DELAYS, WEIGHTS)
from spynnaker.pyNN.utilities import utility_calls
from spynnaker.pyNN.exceptions import SpynnakerException

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)
    from spynnaker.pyNN.models.projection import Projection
    from spynnaker.pyNN.models.populations import Population, PopulationView

# global objects
logger = FormatAdapter(logging.getLogger(__name__))
_expr_context = SafeEval(
    math, numpy, numpy.arccos, numpy.arcsin, numpy.arctan, numpy.arctan2,
    numpy.ceil, numpy.cos, numpy.cosh, numpy.exp, numpy.fabs, numpy.floor,
    numpy.fmod, numpy.hypot, numpy.ldexp, numpy.log, numpy.log10, numpy.modf,
    numpy.power, numpy.sin, numpy.sinh, numpy.sqrt, numpy.tan, numpy.tanh,
    numpy.maximum, numpy.minimum, e=numpy.e, pi=numpy.pi)


class AbstractConnector(object, metaclass=AbstractBase):
    """
    Abstract class that all PyNN Connectors extend.
    """

    NUMPY_SYNAPSES_DTYPE = numpy.dtype(
        [("source", uint32), ("target", uint16),
         ("weight", float64), ("delay", float64),
         ("synapse_type", uint8)])

    __slots__ = (
        "__min_delay",
        "__n_clipped_delays",
        "__safe",
        "__space",
        "__verbose",
        "__param_seeds")

    def __init__(self, safe: bool = True, callback: None = None,
                 verbose: bool = False):
        """
        :param safe: if True, check that weights and delays have valid
            values. If False, this check is skipped. (NB: SpiNNaker always
            checks.)
        :param callback:
            if given, a callable that display a progress bar on the terminal.

            .. note::
                Not supported by sPyNNaker.
        :param verbose:
            Whether to output extra information about the connectivity to a
            CSV file
        """
        if callback is not None:
            warn_once(logger, "sPyNNaker ignores connector callbacks.")
        self.__safe = safe
        self.__space: Optional[Space] = None
        self.__verbose = verbose

        self.__n_clipped_delays = numpy.int64(0)
        self.__min_delay = 0.0
        self.__param_seeds: Dict[Tuple[int, int], int] = dict()

    def set_space(self, space: Space) -> None:
        """
        Set the space object (allowed after instantiation).

        :param space:
        """
        self.__space = space

    def set_projection_information(
            self, synapse_info: SynapseInformation) -> None:
        """
        Sets a connectors projection info.

        :param synapse_info: the synapse info
        """
        _ = synapse_info
        self.__min_delay = SpynnakerDataView.get_simulation_time_step_ms()

    def _get_delay_minimum(
            self, delays: DELAYS, n_connections: int,
            synapse_info: SynapseInformation) -> float:
        """
        Get the minimum delay given a float or RandomDistribution.

        :param delays: the delays
        :param n_connections: how many connections
        """
        if isinstance(delays, RandomDistribution):
            low_estimated_delay = utility_calls.get_minimum_probable_value(
                delays, n_connections)
            low = utility_calls.low(delays)
            if low is None:
                return low_estimated_delay

            # The minimum is the maximum of the possible maximums
            return max(low_estimated_delay, low, 1)
        elif isinstance(delays, str):
            d = self._get_distances(delays, synapse_info)
            return numpy.min(_expr_context.eval(delays, d=d))
        elif is_scalar(delays):
            return delays
        raise self.delay_type_exception(delays)

    def _get_delay_maximum(
            self, delays: DELAYS, n_connections: int,
            synapse_info: SynapseInformation) -> float:
        """
        Get the maximum delay given a float or RandomDistribution.
        """
        if isinstance(delays, RandomDistribution):
            max_estimated_delay = utility_calls.get_maximum_probable_value(
                delays, n_connections)
            high = utility_calls.high(delays)
            if high is None:
                return max_estimated_delay

            # The maximum is the minimum of the possible maximums
            return min(max_estimated_delay, high)
        elif isinstance(delays, str):
            d = self._get_distances(delays, synapse_info)
            return numpy.max(_expr_context.eval(delays, d=d))
        elif is_scalar(delays):
            return delays
        raise self.delay_type_exception(delays)

    @abstractmethod
    def get_delay_maximum(
            self, synapse_info: SynapseInformation) -> float:
        """
        :param synapse_info: the synapse info
        :returns: The maximum delay specified by the user in ms,.
        """
        raise NotImplementedError

    @abstractmethod
    def get_delay_minimum(
            self, synapse_info: SynapseInformation) -> Optional[float]:
        """
        :param synapse_info:
        :returns: The minimum delay specified by the user in ms,
           or `None` if unbounded.
        """
        raise NotImplementedError

    def get_delay_variance(self, delays: DELAYS,
                           synapse_info: SynapseInformation) -> float:
        """
        :param delays:
        :param synapse_info: Info to get distances from (if needed)
        :returns: The variance of the delays.
        """
        if isinstance(delays, RandomDistribution):
            return utility_calls.get_variance(delays)
        elif isinstance(delays, str):
            d = self._get_distances(delays, synapse_info)
            return numpy.var(_expr_context.eval(delays, d=d))
        elif is_scalar(delays):
            return 0.0
        raise self.delay_type_exception(delays)

    def _get_n_connections_from_pre_vertex_with_delay_maximum(
            self, delays: DELAYS, n_total_connections: int,
            n_connections: int, min_delay: float, max_delay: float,
            synapse_info: SynapseInformation) -> int:
        """
        :param delays:
        :param n_total_connections:
        :param n_connections:
        :param min_delay:
        :param max_delay:
        :returns: The expected number of delays
           that will fall within min_delay andmax_delay
           given a float, RandomDistribution or list of delays.
        """
        if isinstance(delays, RandomDistribution):
            prob_in_range = utility_calls.get_probability_within_range(
                delays, min_delay, max_delay)
            if prob_in_range > 0:
                v = int(math.ceil(utility_calls.get_probable_maximum_selected(
                    n_total_connections, n_connections, prob_in_range)))
                # If the probability is so low as to result in 0, assume
                # at least 1 if there is some probability that the delay is
                # in range
                return max(v, 1)
            else:
                return 0
        elif isinstance(delays, str):
            d = self._get_distances(delays, synapse_info)
            delays = _expr_context.eval(delays, d=d)
            n_delayed = sum([len([
                delay for delay in delays
                if min_delay <= delay <= max_delay])])
            if n_delayed == 0:
                return 0
            n_total = len(delays)
            prob_delayed = float(n_delayed) / float(n_total)
            return int(math.ceil(utility_calls.get_probable_maximum_selected(
                n_total_connections, n_connections, prob_delayed)))
        elif is_scalar(delays):
            if min_delay <= delays <= max_delay:
                return int(math.ceil(n_connections))
            return 0
        raise self.delay_type_exception(delays)

    @abstractmethod
    def get_n_connections_from_pre_vertex_maximum(
            self, n_post_atoms: int, synapse_info: SynapseInformation,
            min_delay: Optional[float] = None,
            max_delay: Optional[float] = None) -> int:
        """
        Get the maximum number of connections from any
        neuron in the pre vertex to the neurons in the post_vertex_slice,
        for connections with a delay between min_delay and max_delay
        (inclusive) if both specified (otherwise all connections).

        Not all concrete connectors support omitting the delay range.

        :param n_post_atoms:
        :param synapse_info:
        :param min_delay:
        :param max_delay:
        :returns: maximum number of connections
        """
        raise NotImplementedError

    @abstractmethod
    def get_n_connections_to_post_vertex_maximum(
            self, synapse_info: SynapseInformation) -> int:
        """
        :param synapse_info:
        :returns: The maximum number of connections to any neuron
        in the post vertex from neurons in the pre vertex.
        """
        raise NotImplementedError

    def get_weight_mean(self, weights: WEIGHTS,
                        synapse_info: SynapseInformation) -> float:
        """
        :param weights:
        :param synapse_info: Info to get distances from (if needed)
        :returns: The mean of the weights.
        """
        if isinstance(weights, RandomDistribution):
            return abs(utility_calls.get_mean(weights))
        elif isinstance(weights, str):
            d = self._get_distances(weights, synapse_info)
            return numpy.mean(_expr_context.eval(weights, d=d))
        elif is_scalar(weights):
            return abs(weights)
        raise self.weight_type_exception(synapse_info)

    def _get_weight_maximum(
            self, weights: WEIGHTS, n_connections: int,
            synapse_info: SynapseInformation) -> float:
        """
        Get the maximum of the weights.
        """
        if isinstance(weights, RandomDistribution):
            mean_weight = utility_calls.get_mean(weights)
            if mean_weight < 0:
                min_weight = utility_calls.get_minimum_probable_value(
                    weights, n_connections)
                low = utility_calls.low(weights)
                if low is None:
                    return abs(min_weight)
                return abs(max(min_weight, low))
            else:
                max_weight = utility_calls.get_maximum_probable_value(
                    weights, n_connections)
                high = utility_calls.high(weights)
                if high is None:
                    return abs(max_weight)
                return abs(min(max_weight, high))
        elif isinstance(weights, str):
            d = self._get_distances(weights, synapse_info)
            return numpy.max(_expr_context.eval(weights, d=d))
        elif is_scalar(weights):
            return abs(weights)
        raise self.weight_type_exception(weights)

    @abstractmethod
    def get_weight_maximum(self, synapse_info: SynapseInformation) -> float:
        """
        :param synapse_info:
        :returns: The maximum of the weights for this connection.
        """
        raise NotImplementedError

    def get_weight_variance(self, weights: WEIGHTS,
                            synapse_info: SynapseInformation) -> float:
        """
        :param weights:
        :param synapse_info: Info to get distances from (if needed)
        :returns: The variance of the weights.
        """
        if isinstance(weights, RandomDistribution):
            return utility_calls.get_variance(weights)
        elif isinstance(weights, str):
            d = self._get_distances(weights, synapse_info)
            return numpy.var(_expr_context.eval(weights, d=d))
        elif is_scalar(weights):
            return 0.0
        raise self.weight_type_exception(weights)

    def _expand_distances(self, d_expression: str) -> bool:
        """
        Check if a distance expression contains at least one term `d[x]`.
        If yes, then the distances are expanded to distances in the
        separate coordinates rather than the overall distance over all
        coordinates, and we assume the user has specified an expression
        such as `d[0] + d[2]`.

        :param d_expression:
        """
        regexpr = re.compile(r'.*d\[\d*\].*')
        return bool(regexpr.match(d_expression))

    def _get_distances(self, values: str,
                       synapse_info: SynapseInformation) -> NDArray[float64]:
        if self.__space is None:
            raise self._no_space_exception(values, synapse_info)
        expand_distances = self._expand_distances(values)

        return self.__space.distances(
            synapse_info.pre_population.positions,
            synapse_info.post_population.positions,
            expand_distances)

    def _generate_random_values(
            self, values: RandomDistribution, n_connections: int,
            post_vertex_slice: Slice) -> NDArray[float64]:
        """
        :param values:
        :param n_connections:
        :param post_vertex_slice:
        """
        key = (id(post_vertex_slice), id(values))
        seed = self.__param_seeds.get(key, None)
        if seed is None:
            seed = int(values.rng.next() * 0x7FFFFFFF)
            self.__param_seeds[key] = seed
        new_rng = NumpyRNG(seed)
        copy_rd = RandomDistribution(
            values.name, parameters_pos=None, rng=new_rng,
            **values.parameters)
        if n_connections == 1:
            return numpy.array([copy_rd.next(1)], dtype=float64)
        return copy_rd.next(n_connections)

    def _no_space_exception(
            self, values: WEIGHTS_DELAYS,
            synapse_info: SynapseInformation) -> SpynnakerException:
        """
        :param values:
        :param synapse_info:
        :returns: A SpynnakerException about there being no space defined
        """
        return SpynnakerException(
            f"Str Weights or delays {values} are distance-dependent "
            f"but no space object was specified in projection "
            f"{synapse_info.pre_population}-"
            f"{synapse_info.post_population}")

    def weight_type_exception(
            self, weights: WEIGHTS) -> SpynnakerException:
        """
        :param weights:
        :returns: An Exception explaining incorrect weight or delay type
        """
        if weights is None:
            return SpynnakerException(
                f"The Synapse used is not is not supported with a "
                f"{(type(self))} as neither provided weights")
        elif isinstance(weights, str):
            return SpynnakerException(
                f"Str Weights {weights} not supported by a {(type(self))}")
        elif isinstance(weights, numpy.ndarray):
            # The problem is that these methods are for a MachineVertex/ core
            # while weight and delay are supplied at the application level
            # The FromList is also the one designed to handle the 2D case
            return SpynnakerException(
                f"For efficiency reason {type(self)} does not supports "
                f"list or arrays for weight."
                f"Please use a FromListConnector instead")
        else:
            return SpynnakerException(f"Unrecognised weight {weights}")

    def delay_type_exception(self, delays: DELAYS) -> SpynnakerException:
        """
        :param delays:
        :returns: An Exception explaining incorrect delay type
        """
        if isinstance(delays, str):
            return SpynnakerException(
                f"Str delays {delays} not supported by {(type(self))}")
        elif isinstance(delays, numpy.ndarray):
            # The problem is that these methods are for a MachineVertex/ core
            # while weight and delay are supplied at the application level
            # The FromList is also the one designed to handle the 2D case
            return SpynnakerException(
                f"For efficiency reason {type(self)} does not supports "
                f"list or arrays for weight or delay."
                f"Please use a FromListConnector instead")
        else:
            return SpynnakerException(f"Unrecognised delay {delays}")

    def _generate_values(
            self, values: WEIGHTS_DELAYS, sources: numpy.ndarray,
            targets: numpy.ndarray, n_connections: int, post_slice: Slice,
            synapse_info: SynapseInformation,
            weights: bool) -> NDArray[float64]:
        if isinstance(values, RandomDistribution):
            return self._generate_random_values(
                values, n_connections, post_slice)
        elif isinstance(values, str) or callable(values):
            if self.__space is None:
                raise self._no_space_exception(values, synapse_info)

            expand_distances = True
            if isinstance(values, str):
                expand_distances = self._expand_distances(values)

                # At this point we need to now get the values corresponding to
                # the distances between connections in "sources" and "targets"
                eval_values = numpy.zeros(n_connections, dtype=float64)
                for i in range(n_connections):
                    # get the distance for this source and target pair
                    dist = self.__space.distances(
                        synapse_info.pre_population.positions[sources[i]],
                        synapse_info.post_population.positions[targets[i]],
                        expand_distances)
                    # evaluate expression at this distance
                    eval_values[i] = _expr_context.eval(values, d=dist)
                return eval_values

            d = self.__space.distances(
                synapse_info.pre_population.positions[sources[i]],
                synapse_info.post_population.positions[targets[i]],
                expand_distances)

            return values(d)
        elif is_scalar(values):
            return numpy.repeat([values], n_connections).astype(float64)
        if weights:
            raise self.weight_type_exception(values)
        else:
            raise self.delay_type_exception(values)

    def _generate_weights(
            self, sources: numpy.ndarray, targets: numpy.ndarray,
            n_connections: int, post_slice: Slice,
            synapse_info: SynapseInformation) -> numpy.ndarray:
        """
        Generate weight values.
        """
        weights = self._generate_values(
            synapse_info.weights, sources, targets, n_connections, post_slice,
            synapse_info, weights=True)
        if self.__safe:
            if not weights.size:
                warn_once(logger, "No connection in " + str(self))
            elif numpy.amin(weights) < 0 < numpy.amax(weights):
                raise SpynnakerException(
                    "Weights must be either all positive or all negative in "
                    f"projection {synapse_info.pre_population.label}->"
                    f"{synapse_info.post_population.label}")
        return numpy.abs(weights)

    def _clip_delays(self, delays: NDArray[float64]) -> NDArray[float64]:
        """
        Clip delay values, keeping track of how many have been clipped.
        """
        # count values that could be clipped
        self.__n_clipped_delays = numpy.sum(delays < self.__min_delay)

        # clip values
        if delays.size:
            delays[delays < self.__min_delay] = self.__min_delay
        return delays

    def _generate_delays(
            self, sources: numpy.ndarray, targets: numpy.ndarray,
            n_connections: int, post_slice: Slice,
            synapse_info: SynapseInformation) -> numpy.ndarray:
        """
        Generate valid delay values.
        """
        delays = self._generate_values(
            synapse_info.delays, sources, targets, n_connections, post_slice,
            synapse_info, weights=False)

        return self._clip_delays(delays)

    @staticmethod
    def __pop_label(pop: Union[Population, PopulationView]) -> str:
        lbl = pop.label
        if lbl is None:
            raise ValueError("unlabelled population")
        return lbl

    def get_provenance_data(self, synapse_info: SynapseInformation) -> None:
        """
        Adds the synapse information to provenance database.

        :param synapse_info:
        """
        # Convert to native Python integer; provenance system assumption
        ncd = self.__n_clipped_delays.item()
        with ProvenanceWriter() as db:
            db.insert_connector(
                self.__pop_label(synapse_info.pre_population),
                self.__pop_label(synapse_info.post_population),
                self.__class__.__name__, "Times_synaptic_delays_got_clipped",
                ncd)
            if ncd > 0:
                db.insert_report(
                    f"The delays in the connector {self.__class__.__name__} "
                    f"from {synapse_info.pre_population.label} "
                    f"to {synapse_info.post_population.label} "
                    f"was clipped to {self.__min_delay} a total of {ncd} "
                    f"times. This can be avoided by reducing the timestep or "
                    f"increasing the minimum delay to one timestep")

    @property
    def safe(self) -> bool:
        """
        Safe value supplied. Note ignored by sPyNNaker
        """
        return self.__safe

    @safe.setter
    def safe(self, new_value: bool) -> None:
        self.__safe = new_value

    @property
    def space(self) -> Optional[Space]:
        """
        The space object (may be updated after instantiation).
        """
        return self.__space

    @space.setter
    def space(self, new_value: Space) -> None:
        """
        Set the space object (allowed after instantiation).

        :param new_value:
        """
        self.__space = new_value

    @property
    def verbose(self) -> bool:
        """
        verbose value supplied by user
        """
        return self.__verbose

    @verbose.setter
    def verbose(self, new_value: bool) -> None:
        self.__verbose = new_value

    def get_connected_vertices(
            self, s_info: SynapseInformation, source_vertex: ApplicationVertex,
            target_vertex: ApplicationVertex) -> Sequence[
                Tuple[MachineVertex, Sequence[AbstractVertex]]]:
        """
        Get the machine vertices that are connected to each other with
        this connector

        :param s_info: The synapse information of the connection
        :param source_vertex: The source of the spikes
        :param target_vertex: The target of the spikes
        :return: A list of tuples of (target machine vertex, list of sources)
        """
        # By default, just return that the whole target connects to the
        # whole source
        return [(m_vertex, [source_vertex])
                for m_vertex in target_vertex.splitter.get_in_coming_vertices(
                    s_info.partition_id)]

    def connect(self, projection: Projection) -> Never:
        """
        Apply this connector to a projection.

        .. warning::
            Do *not* call this! SpyNNaker does not work that way.

        :param projection:
        :raises SpynnakerException: Always. Method not supported; profiled out.
        """
        raise SpynnakerException("Standard pyNN connect method not supported")

    @staticmethod
    def _roundsize(size: Union[int, float], label: str) -> int:
        """
        Ensures that the ``size`` is an integer. Approximate integers are
        rounded; other values cause exceptions.

        :param size: The value to be rounded
        :param label: The type-name of the connection, for messages
        :raises SpynnakerException: If the size is non-integer and not close
        """
        if isinstance(size, int):
            return size
        # Allow a float which has a near int value
        temp = int(round(size))
        if abs(temp - size) < 0.001:
            logger.warning("Size of {} rounded from {} to {}. "
                           "Please use int values for size",
                           label, size, temp)
            return temp
        raise SpynnakerException(
            f"Size of {label} must be an int, received {size}")

    def validate_connection(
            self, application_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> None:
        """
        Checks that the edge supports the connector.  Returns nothing; it
        is assumed that an Exception will be raised if anything is wrong.

        By default this checks only that the views are not used
        on multi-dimensional vertices.

        :param application_edge: The edge of the connection
        :param synapse_info: The synaptic information
        """
        _ = application_edge
        if ((synapse_info.prepop_is_view and
                len(synapse_info.pre_vertex.atoms_shape) > 1) or
                (synapse_info.postpop_is_view and
                 len(synapse_info.post_vertex.atoms_shape) > 1)):
            raise ConfigurationException(
                "Using a projection where the source or target is a "
                "PopulationView on a multi-dimensional Population is not "
                "supported")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from pacman.model.graphs.application import ApplicationSpiNNakerLinkVertex
from spynnaker.pyNN.external_devices_models.push_bot.parameters import (
    PushBotMotor)
from spynnaker.pyNN.external_devices_models.push_bot.ethernet import (
    PushBotEthernetMotorDevice)
from spynnaker.pyNN.models.common import PopulationApplicationVertex
from spynnaker.pyNN.protocols import MunichIoSpiNNakerLinkProtocol


class PushBotSpiNNakerLinkMotorDevice(
        PushBotEthernetMotorDevice, ApplicationSpiNNakerLinkVertex,
        PopulationApplicationVertex):
    """
    The motor of a PushBot.
    """
    __slots__ = ()

    def __init__(
            self, motor: PushBotMotor, protocol: MunichIoSpiNNakerLinkProtocol,
            spinnaker_link_id: int,  n_neurons: int = 1,
            label: Optional[str] = None,  board_address: Optional[str] = None):
        """
        :param motor: the motor to control
        :param protocol: The protocol used to control the device
        :param spinnaker_link_id: The SpiNNakerLink connected to
        :param n_neurons: The number of neurons in the device
        :param label: The label of the device
        :param board_address:
            The IP address of the board that the device is connected to
        """
        super().__init__(motor, protocol)
        ApplicationSpiNNakerLinkVertex.__init__(
            self, spinnaker_link_id=spinnaker_link_id, n_atoms=n_neurons,
            board_address=board_address, label=label)



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Iterable
from spinn_utilities.abstract_base import AbstractBase, abstractmethod


class AbstractHasParameterNames(object, metaclass=AbstractBase):
    """
    A component that has parameter names. Parameter names are usually
    properties of the component, and are frequently also settable by named
    parameter when making the component.
    """

    __slots__ = ()

    @abstractmethod
    def get_parameter_names(self) -> Iterable[str]:
        """
        :returns: The parameter names available from the component.
        """
        raise NotImplementedError



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import List, Sequence, Tuple, TYPE_CHECKING, cast
from spinn_utilities.overrides import overrides
from spinn_utilities.config_holder import get_config_bool
from pacman.model.graphs.application import (
    ApplicationEdgePartition, ApplicationVertex)
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from spynnaker.pyNN.models.abstract_models import ColouredApplicationVertex
from spynnaker.pyNN.exceptions import DelayExtensionException
from spynnaker.pyNN.models.abstract_models import AbstractHasDelayStages
from spynnaker.pyNN.utilities.constants import POP_TABLE_MAX_ROW_LENGTH
from spynnaker.pyNN.models.neural_projections import DelayedApplicationEdge
if TYPE_CHECKING:
    from spynnaker.pyNN.extra_algorithms.splitter_components import (
        SplitterDelayVertexSlice)

_DELAY_PARAM_HEADER_WORDS = 9


class DelayExtensionVertex(ColouredApplicationVertex, AbstractHasDelayStages):
    """
    Provide delays to incoming spikes in multiples of the maximum delays
    of a neuron (typically 16 or 32).
    """
    __slots__ = (
        # The partition this Delay is supporting
        "__partition",
        "__delay_per_stage",
        "__n_delay_stages",
        "__drop_late_spikes",
        "__outgoing_edges",
        "__n_colour_bits")

    # this maps to what master assumes
    MAX_SLOTS = 8
    SAFETY_FACTOR = 5000
    MAX_DTCM_AVAILABLE = 59756 - SAFETY_FACTOR

    def __init__(
            self, partition: ApplicationEdgePartition, delay_per_stage: int,
            n_delay_stages: int, n_colour_bits: int,
            label: str = "DelayExtension"):
        """
        :param partition: The partition that this delay is supporting
        :param delay_per_stage: the delay per stage
        :param n_delay_stages: the (initial) number of delay stages needed
        :param n_colour_bits: the number of bits for event colouring
        :param label: the vertex label
        """
        super().__init__(
            label, POP_TABLE_MAX_ROW_LENGTH, splitter=None)

        self.__partition = partition
        self.__n_delay_stages = n_delay_stages
        self.__delay_per_stage = delay_per_stage

        self.__drop_late_spikes = get_config_bool(
            "Simulation", "drop_late_spikes") or False

        self.__outgoing_edges: List[DelayedApplicationEdge] = list()

        self.__n_colour_bits = n_colour_bits

    @property
    def n_atoms(self) -> int:
        """
        The number of atoms in this vertex.
        """
        return self.__partition.pre_vertex.n_atoms

    @property
    @overrides(ColouredApplicationVertex.atoms_shape)
    def atoms_shape(self) -> Tuple[int, ...]:
        return self.__partition.pre_vertex.atoms_shape

    @property
    def _delay_splitter(self) -> SplitterDelayVertexSlice:
        if TYPE_CHECKING:
            return cast(SplitterDelayVertexSlice, self._splitter)
        else:
            return self._splitter

    @property
    def drop_late_spikes(self) -> bool:
        """
        Whether to drop late spikes.
        """
        return self.__drop_late_spikes

    @staticmethod
    def get_max_delay_ticks_supported(delay_ticks_at_post_vertex: int) -> int:
        """
        :param delay_ticks_at_post_vertex:
        :returns:
           The max ticks the combination of delay and post vertex can handle.
        """
        return DelayExtensionVertex.MAX_SLOTS * delay_ticks_at_post_vertex

    @property
    @overrides(AbstractHasDelayStages.n_delay_stages)
    def n_delay_stages(self) -> int:
        return self.__n_delay_stages

    def set_new_n_delay_stages_and_delay_per_stage(
            self, n_delay_stages: int, delay_per_stage: int) -> None:
        """
        Sets the delays per stages. Verifies delay per stage is unchanged.

        :param n_delay_stages:
        :param delay_per_stage:
        """
        if delay_per_stage != self.__delay_per_stage:
            raise DelayExtensionException(
                "The delay per stage is already set to "
                f"{self.__delay_per_stage}, and therefore {delay_per_stage} "
                "is not yet feasible. "
                "Please report it to Spinnaker user mail list.")

        if n_delay_stages > self.__n_delay_stages:
            self.__n_delay_stages = n_delay_stages

    @property
    def delay_per_stage(self) -> int:
        """
        The delay per stage, in timesteps.
        """
        return self.__delay_per_stage

    @property
    def source_vertex(self) -> ApplicationVertex:
        """
        The pre vertex of the delayed partition.
        """
        return self.__partition.pre_vertex

    def delay_params_size(self) -> int:
        """
        :returns: The size of the delay parameters.
        """
        return BYTES_PER_WORD * _DELAY_PARAM_HEADER_WORDS

    @property
    def partition(self) -> ApplicationEdgePartition:
        """
        The partition that this delay is supporting.
        """
        return self.__partition

    def add_outgoing_edge(self, edge: DelayedApplicationEdge) -> None:
        """
        Add an outgoing edge to the delay extension.

        :param edge: The edge to add
        """
        self.__outgoing_edges.append(edge)

    @property
    def outgoing_edges(self) -> Sequence[DelayedApplicationEdge]:
        """
        The outgoing edges from this vertex.
        """
        return self.__outgoing_edges

    @property
    @overrides(ColouredApplicationVertex.n_colour_bits)
    def n_colour_bits(self) -> int:
        return self.__n_colour_bits



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import cast, Final

import numpy
from numpy.typing import NDArray

from spynnaker.pyNN.utilities import utility_calls

from .spike_source_array import SpikeSourceArray

_inf: Final = float('inf')


class SpikeSourceFromFile(SpikeSourceArray):
    """
    A spike source that works from a file (typically a tab-separated table in
    a text file).
    """

    def __init__(
            self, spike_time_file: str,
            min_atom: float = 0.0, max_atom: float = _inf,
            min_time: float = 0.0, max_time: float = _inf,
            split_value: str = "\t"):
        """
        :param spike_time_file:
            absolute path to a file containing spike values
        :param min_atom: min neuron ID to which neurons to read in
        :param max_atom: max neuron ID to which neurons to read in
        :param min_time: min time slot to read neurons values of.
        :param max_time: max time slot to read neurons values of.
        :param split_value: the pattern to split by
        """
        spike_times = utility_calls.read_spikes_from_file(
            spike_time_file, min_atom, max_atom, min_time, max_time,
            split_value)
        super().__init__(spike_times)

    @property
    def spike_times(self) -> NDArray[numpy.integer]:
        """
        The spike times read from the file.
        """
        # type guaranteed by init
        return cast(NDArray[numpy.integer], self._spike_times)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List, Tuple

from numpy import integer, uint32
from numpy.typing import NDArray

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
    ConnectionsArray)
from .abstract_sdram_synapse_dynamics import AbstractSDRAMSynapseDynamics


class AbstractStaticSynapseDynamics(
        AbstractSDRAMSynapseDynamics, metaclass=AbstractBase):
    """
    Dynamics which don't change over time.
    """

    __slots__ = ()

    @abstractmethod
    def get_n_words_for_static_connections(self, n_connections: int) -> int:
        """
        :param n_connections:
        :returns:
           The number of 32-bit words for `n_connections` in a single row.
        """
        raise NotImplementedError

    @abstractmethod
    def get_static_synaptic_data(
            self, connections: ConnectionsArray,
            connection_row_indices: NDArray[integer], n_rows: int,
            n_synapse_types: int,
            max_n_synapses: int, max_atoms_per_core: int) -> Tuple[
                List[NDArray[uint32]], NDArray[integer]]:
        """
        Get the fixed-fixed data for each row, and lengths for the
        fixed-fixed parts of each row.

        Data is returned as an array made up of an array of 32-bit words for
        each row for the fixed-fixed region. The row into which connection
        should go is given by `connection_row_indices`, and the total number
        of rows is given by `n_rows`.

        Lengths are returned as an array made up of an integer for each row,
        for the fixed-fixed region.

        :param connections: The connections to get data for
        :param connection_row_indices:
            The row into which each connection should go
        :param n_rows: The number of rows to write
        :param n_synapse_types: The number of synapse types
        :param max_n_synapses: The maximum number of synapses to generate
        :param max_atoms_per_core: The maximum number of atoms on a core
        :return: (ff_data, ff_size)
        """
        raise NotImplementedError

    @abstractmethod
    def get_n_static_words_per_row(
            self, ff_size: NDArray[integer]) -> NDArray[integer]:
        """
        :param ff_size:
        :returns: The number of bytes to be read per row for the static data
           given the size that was written to each row.
        """
        raise NotImplementedError

    @abstractmethod
    def get_n_synapses_in_rows(
            self, ff_size: NDArray[integer]) -> NDArray[integer]:
        """
        :param ff_size:
        :returns: The number of synapses in the rows with sizes `ff_size`.
        """
        raise NotImplementedError

    @abstractmethod
    def read_static_synaptic_data(
            self, n_synapse_types: int,
            ff_size: NDArray[integer], ff_data: List[NDArray[uint32]],
            max_atoms_per_core: int) -> ConnectionsArray:
        """
        Read the connections from the words of data in `ff_data`.

        :param n_synapse_types:
        :param ff_size:
        :param ff_data:
        :param max_atoms_per_core:
        :return: the connections read with dtype
            :py:const:`~.NUMPY_CONNECTORS_DTYPE`
        """
        raise NotImplementedError



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Iterable

from numpy import floating
from numpy.typing import NDArray

from spinn_utilities.overrides import overrides

from spinn_front_end_common.interface.ds import (
    DataType, DataSpecificationBase)
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from .abstract_has_a_plus_a_minus import AbstractHasAPlusAMinus
from .abstract_weight_dependence import AbstractWeightDependence

# Four words per synapse type
_SPACE_PER_SYNAPSE_TYPE = 4 * BYTES_PER_WORD


class WeightDependenceAdditive(
        AbstractHasAPlusAMinus, AbstractWeightDependence):
    """
    An additive weight dependence STDP rule.
    """

    __slots__ = (
        "__w_max",
        "__w_min")
    __PARAM_NAMES = ('w_min', 'w_max', 'A_plus', 'A_minus')

    # noinspection PyPep8Naming
    def __init__(self, w_min: float = 0.0, w_max: float = 1.0):
        """
        :param w_min: :math:`w^{min}`
        :param w_max: :math:`w^{max}`
        """
        super().__init__()
        self.__w_min = w_min
        self.__w_max = w_max

    @property
    def w_min(self) -> float:
        """
        :math:`w^{min}`
        """
        return self.__w_min

    @property
    def w_max(self) -> float:
        """
        :math:`w^{max}`
        """
        return self.__w_max

    @overrides(AbstractWeightDependence.is_same_as)
    def is_same_as(self, weight_dependence: AbstractWeightDependence) -> bool:
        if not isinstance(weight_dependence, WeightDependenceAdditive):
            return False
        return (
            (self.__w_min == weight_dependence.w_min) and
            (self.__w_max == weight_dependence.w_max) and
            (self.A_plus == weight_dependence.A_plus) and
            (self.A_minus == weight_dependence.A_minus))

    @property
    def vertex_executable_suffix(self) -> str:
        """
        The suffix to be appended to the vertex executable for this rule.
        """
        return "additive"

    @overrides(AbstractWeightDependence.get_parameters_sdram_usage_in_bytes)
    def get_parameters_sdram_usage_in_bytes(
            self, n_synapse_types: int, n_weight_terms: int) -> int:
        if n_weight_terms != 1:
            raise NotImplementedError(
                "Additive weight dependence only supports one term")
        return _SPACE_PER_SYNAPSE_TYPE * n_synapse_types

    @overrides(AbstractWeightDependence.write_parameters)
    def write_parameters(
            self, spec: DataSpecificationBase, global_weight_scale: float,
            synapse_weight_scales: NDArray[floating],
            n_weight_terms: int) -> None:
        # Loop through each synapse type
        for _ in synapse_weight_scales:
            # Scale the weights
            spec.write_value(
                data=self.__w_min * global_weight_scale,
                data_type=DataType.S1615)
            spec.write_value(
                data=self.__w_max * global_weight_scale,
                data_type=DataType.S1615)

            # pylint: disable=wrong-spelling-in-comment
            # Based on http://data.andrewdavison.info/docs/PyNN/_modules/pyNN
            #                /standardmodels/synapses.html
            # Pre-multiply A+ and A- by Wmax
            spec.write_value(
                data=self.A_plus * global_weight_scale,
                data_type=DataType.S1615)
            spec.write_value(
                data=self.A_minus * global_weight_scale,
                data_type=DataType.S1615)

    @property
    def weight_maximum(self) -> float:
        """
        The maximum weight that will ever be set in a synapse as a result
        of this rule.
        """
        return self.__w_max

    @overrides(AbstractWeightDependence.get_parameter_names)
    def get_parameter_names(self) -> Iterable[str]:
        return self.__PARAM_NAMES



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Set

from spinn_utilities.config_holder import (
    clear_cfg_files, set_cfg_files)
from spinn_front_end_common.interface.config_setup import (
    add_default_cfg, add_spinnaker_cfg)
from spinn_front_end_common.interface.config_setup import (
    fec_cfg_paths_skipped)

from spynnaker.pyNN.data.spynnaker_data_writer import SpynnakerDataWriter

CONFIG_FILE_NAME = "spynnaker.cfg"


def setup_configs() -> None:
    """
    Sets up the configurations including the users configuration file.

    Clears out any previous read configurations but does not load the new
    configurations so a warning is generated if a configuration is used before
    setup is called.
    """
    clear_cfg_files(False)
    add_spinnaker_cfg()  # This add its dependencies too
    set_cfg_files(
        config_file=CONFIG_FILE_NAME,
        default=os.path.join(os.path.dirname(__file__), CONFIG_FILE_NAME))


def unittest_setup() -> None:
    """
    Does all the steps that may be required before a unit-test.

    Resets the configurations so only the local default configurations are
    included.
    The user configuration is *not* included!

    Unsets any previous simulators and temporary directories

    .. note::
         This file should only be called from sPyNNaker tests
         that do not call `sim.setup`
    """
    clear_cfg_files(True)
    add_spynnaker_cfg()
    SpynnakerDataWriter.mock()


def add_spynnaker_cfg() -> None:
    """
    Add the local configuration and all dependent configuration files.
    """
    add_spinnaker_cfg()  # This add its dependencies too
    add_default_cfg(os.path.join(os.path.dirname(__file__), CONFIG_FILE_NAME))


def cfg_paths_skipped() -> Set[str]:
    """
    Set of cfg path that would not be found based on other cfg settings

    Assuming mode = Debug

    :returns:
       List of cfg Option names that point to paths unlikely to be used.
    """
    return fec_cfg_paths_skipped()



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from dataclasses import dataclass
from typing import List, Optional, Tuple, Union, TYPE_CHECKING

import numpy
from numpy import integer, uint32
from numpy.typing import NDArray

from pacman.model.graphs.application import ApplicationVertex
from pacman.model.graphs.common import Slice

from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.models.neural_projections.connectors import (
    AbstractConnector)
from spynnaker.pyNN.exceptions import SynapseRowTooBigException
from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    AbstractStaticSynapseDynamics, AbstractSDRAMSynapseDynamics,
    AbstractPlasticSynapseDynamics)
from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
    NUMPY_CONNECTORS_DTYPE, ConnectionsArray)
from spynnaker.pyNN.types import WeightScales

from .master_pop_table import MasterPopTableAsBinarySearch

if TYPE_CHECKING:
    from typing_extensions import TypeAlias
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)
    from spynnaker.pyNN.models.neuron.synapse_dynamics import (
        AbstractSynapseDynamics)
    _RowData: TypeAlias = numpy.ndarray  # 2D

_N_HEADER_WORDS = 3
# There are 16 slots, one per time step
_STD_DELAY_SLOTS = 16


@dataclass(frozen=True)
class MaxRowInfo(object):
    """
    Information about the maximums for rows in a synaptic matrix.
    """

    #: Maximum number of synapses in a row of the undelayed matrix.
    undelayed_max_n_synapses: int

    #: Maximum number of synapses in a row of the delayed matrix
    delayed_max_n_synapses: int

    #: Maximum number of bytes, including headers, in a row of the
    #: undelayed matrix, or 0 if no synapses.
    undelayed_max_bytes: int

    #: Maximum number of bytes, including headers, in a row of the
    #: delayed matrix, or 0 if no synapses.
    delayed_max_bytes: int

    #: Maximum number of words, excluding headers, in a row of the
    #: undelayed matrix.
    undelayed_max_words: int

    #: Maximum number of words, excluding headers, in a row of the
    #: delayed matrix.
    delayed_max_words: int


def get_maximum_delay_supported_in_ms(
        post_vertex_max_delay_ticks: int) -> float:
    """
    Get the maximum delay supported by the synapse representation
    before extensions are required.

    :param post_vertex_max_delay_ticks: post vertex max delay
    :return: Maximum delay, in milliseconds.
    """
    return (post_vertex_max_delay_ticks *
            SpynnakerDataView.get_simulation_time_step_ms())


def get_max_row_info(
        synapse_info: SynapseInformation, n_post_atoms: int,
        n_delay_stages: int, in_edge: ProjectionApplicationEdge) -> MaxRowInfo:
    """
    Get the information about the maximum lengths of delayed and
    undelayed rows in bytes (including header), words (without header)
    and number of synapses.

    :param synapse_info:
        The synapse information to get the row data for
    :param n_post_atoms:
        The number of post atoms to get the maximum for
    :param n_delay_stages:
        The number of delay stages on the edge
    :param in_edge:
        The incoming edge on which the synapse information is held
    :returns: Information about the maximums for rows in a synaptic matrix.
    :raises SynapseRowTooBigException:
        If the synapse information can't be represented
    """
    max_delay_supported = get_maximum_delay_supported_in_ms(
        in_edge.post_vertex.splitter.max_support_delay())
    max_delay = max_delay_supported * (n_delay_stages + 1)
    pad_to_length: Optional[int] = None
    if isinstance(synapse_info.synapse_dynamics, AbstractSDRAMSynapseDynamics):
        pad_to_length = synapse_info.synapse_dynamics.pad_to_length

    # delay point where delay extensions start
    min_delay_for_delay_extension = (
        max_delay_supported + numpy.finfo(numpy.double).tiny).item()

    # row length for the non-delayed synaptic matrix
    max_undelayed_n_synapses = synapse_info.connector \
        .get_n_connections_from_pre_vertex_maximum(
            n_post_atoms, synapse_info, 0, max_delay_supported)
    if pad_to_length is not None:
        max_undelayed_n_synapses = max(
            pad_to_length, max_undelayed_n_synapses)

    # determine the max row length in the delay extension
    max_delayed_n_synapses = 0
    if n_delay_stages > 0:
        max_delayed_n_synapses = synapse_info.connector \
            .get_n_connections_from_pre_vertex_maximum(
                n_post_atoms, synapse_info,
                min_delay_for_delay_extension, max_delay)
        if pad_to_length is not None:
            max_delayed_n_synapses = max(
                pad_to_length, max_delayed_n_synapses)

    # Get the row sizes
    dynamics = synapse_info.synapse_dynamics
    if isinstance(dynamics, AbstractStaticSynapseDynamics):
        undelayed_n_words = dynamics.get_n_words_for_static_connections(
            max_undelayed_n_synapses)
        delayed_n_words = dynamics.get_n_words_for_static_connections(
            max_delayed_n_synapses)
    elif isinstance(dynamics, AbstractPlasticSynapseDynamics):
        undelayed_n_words = dynamics.get_n_words_for_plastic_connections(
            max_undelayed_n_synapses)
        delayed_n_words = dynamics.get_n_words_for_plastic_connections(
            max_delayed_n_synapses)
    else:
        raise TypeError(f"{dynamics=} has an unexpected type {type(dynamics)}")

    # Adjust for the allowed row lengths from the population table
    undelayed_max_n_words = _get_allowed_row_length(
        undelayed_n_words, dynamics, in_edge,  max_undelayed_n_synapses)
    delayed_max_n_words = _get_allowed_row_length(
        delayed_n_words, dynamics, in_edge, max_delayed_n_synapses)

    undelayed_max_bytes = 0
    if undelayed_max_n_words > 0:
        undelayed_max_bytes = (
            undelayed_max_n_words + _N_HEADER_WORDS) * BYTES_PER_WORD
    delayed_max_bytes = 0
    if delayed_max_n_words > 0:
        delayed_max_bytes = (
            delayed_max_n_words + _N_HEADER_WORDS) * BYTES_PER_WORD

    return MaxRowInfo(
        max_undelayed_n_synapses, max_delayed_n_synapses,
        undelayed_max_bytes, delayed_max_bytes,
        undelayed_max_n_words, delayed_max_n_words)


def _get_allowed_row_length(
        n_words: int, dynamics: AbstractSynapseDynamics,
        in_edge: ProjectionApplicationEdge, n_synapses: int) -> int:
    """
    Get the allowed row length in words in the population table for a
    desired row length in words.

    :param n_words: The number of words in the row
    :param dynamics: The synapse dynamics used
    :param in_edge: The incoming edge
    :param n_synapses: The number of synapses for the number of words
    :raises SynapseRowTooBigException:
        If the given row is too big; the exception will detail the maximum
        number of synapses that are supported.
    """
    if n_words == 0:
        return 0
    try:
        return MasterPopTableAsBinarySearch.get_allowed_row_length(n_words)
    except SynapseRowTooBigException as e:
        # Find the number of synapses available for the maximum population
        # table size, as extracted from the exception
        if isinstance(dynamics, AbstractSDRAMSynapseDynamics):
            max_synapses = dynamics.get_max_synapses(e.max_size)
            raise SynapseRowTooBigException(
                max_synapses,
                f"The connection between {in_edge.pre_vertex} and "
                f"{in_edge.post_vertex} has more synapses ({n_synapses}) than "
                "can currently be supported on this implementation of PyNN "
                f"({max_synapses} for this connection type). "
                "Please reduce the size of the target population, or reduce "
                "the number of neurons per core.") from e
        # Ugh!
        raise


def get_synapses(
        connections: ConnectionsArray, synapse_info: SynapseInformation,
        n_delay_stages: int, n_synapse_types: int,
        weight_scales: WeightScales, app_edge: ProjectionApplicationEdge,
        max_row_info: MaxRowInfo, gen_undelayed: bool, gen_delayed: bool,
        max_atoms_per_core: int) -> Tuple[_RowData, _RowData]:
    """
    Get the synapses as an array of words for non-delayed synapses and
    an array of words for delayed synapses. This is used to prepare
    information for *deployment to SpiNNaker*.

    :param connections:
        The connections to get the synapses from
    :param synapse_info:
        The synapse information to convert to synapses
    :param n_delay_stages:
        The number of delay stages in total to be represented
    :param n_synapse_types:
        The number of synapse types in total to be represented
    :param weight_scales: The scaling of the weights for each synapse type
    :param app_edge:
        The incoming machine edge that the synapses are on
    :param max_row_info:
        The maximum row information for the synapses
    :param gen_undelayed:
        Whether to generate undelayed data
    :param gen_delayed:
        Whether to generate delayed data
    :param max_atoms_per_core:
        The maximum number of atoms on a core
    :return:
        (``row_data``, ``delayed_row_data``) where:

        * ``row_data`` is the undelayed connectivity data arranged into a
            row per source, each row the same length
        * ``delayed_row_data`` is the delayed connectivity data arranged
            into a row per source per delay stage, each row the same length
    """
    # Get delays in timesteps
    max_delay = app_edge.post_vertex.splitter.max_support_delay()

    # Convert delays to timesteps
    connections["delay"] = numpy.rint(
        connections["delay"] *
        SpynnakerDataView.get_simulation_time_step_per_ms())

    # Scale weights
    if not synapse_info.synapse_type_from_dynamics:
        connections["weight"] = (connections["weight"] * weight_scales[
            synapse_info.synapse_type])

    # Split the connections up based on the delays
    if max_delay is not None:
        plastic_delay_mask = (connections["delay"] <= max_delay)
        undelayed_connections = connections[numpy.where(plastic_delay_mask)]
        delayed_connections = connections[numpy.where(~plastic_delay_mask)]
    else:
        undelayed_connections = connections
        delayed_connections = numpy.zeros(
            0, dtype=AbstractConnector.NUMPY_SYNAPSES_DTYPE)

    # Get the data for the connections
    row_data = numpy.zeros(0, dtype=uint32)
    if gen_undelayed and max_row_info.undelayed_max_n_synapses:
        # Get which row each connection will go into
        undelayed_row_indices = undelayed_connections["source"]
        row_data = _get_row_data(
            undelayed_connections, undelayed_row_indices,
            app_edge.pre_vertex.n_atoms, n_synapse_types,
            synapse_info.synapse_dynamics,
            max_row_info.undelayed_max_n_synapses,
            max_row_info.undelayed_max_words, max_atoms_per_core)
        del undelayed_row_indices
    del undelayed_connections

    # Get the data for the delayed connections
    delayed_row_data = numpy.zeros(0, dtype=uint32)
    if gen_delayed and max_row_info.delayed_max_n_synapses:
        # Get the delay stages and which row each delayed connection will
        # go into
        stages = numpy.floor((numpy.round(
            delayed_connections["delay"] - 1.0)) / max_delay).astype(uint32)
        delayed_cores = delayed_connections["source"] // max_atoms_per_core
        local_sources = delayed_connections["source"] - (
            delayed_cores * max_atoms_per_core)
        delay_atoms_per_core = max_atoms_per_core * n_delay_stages
        delayed_pre_cores = delayed_cores * delay_atoms_per_core
        local_index = ((stages - 1) * max_atoms_per_core) + local_sources
        delayed_row_indices = delayed_pre_cores + local_index
        delayed_connections["delay"] -= max_delay * stages

        # Get the data
        delayed_row_data = _get_row_data(
            delayed_connections, delayed_row_indices,
            app_edge.pre_vertex.n_atoms * n_delay_stages,
            n_synapse_types, synapse_info.synapse_dynamics,
            max_row_info.delayed_max_n_synapses,
            max_row_info.delayed_max_words, max_atoms_per_core)
        del delayed_row_indices
    del delayed_connections

    return row_data, delayed_row_data


def _get_row_data(
        connections: ConnectionsArray, row_indices: NDArray[numpy.integer],
        n_rows: int, n_synapse_types: int,
        synapse_dynamics: AbstractSynapseDynamics, max_row_n_synapses: int,
        max_row_n_words: int, max_atoms_per_core: int) -> _RowData:
    """
    :param connections:
        The connections to convert; the dtype is
        AbstractConnector.NUMPY_SYNAPSES_DTYPE
    :param row_indices:
        The row into which each connection should go; same length as
        connections
    :param n_rows: The total number of rows
    :param n_synapse_types: The number of synapse types allowed
    :param synapse_dynamics:
        The synapse dynamics of the synapses
    :param max_row_n_synapses: The maximum number of synapses in a row
    :param max_row_n_words: The maximum number of words in a row
    :param max_atoms_per_core: The maximum number of atoms per core
    """
    fp_data: Union[NDArray[uint32], List[NDArray[uint32]]]
    pp_data: Union[NDArray[uint32], List[NDArray[uint32]]]
    if isinstance(synapse_dynamics, AbstractStaticSynapseDynamics):
        # Get the static data
        ff_data, ff_size = synapse_dynamics.get_static_synaptic_data(
            connections, row_indices, n_rows, n_synapse_types,
            max_row_n_synapses, max_atoms_per_core)

        # Blank the plastic data
        fp_data = numpy.zeros((n_rows, 0), dtype=uint32)
        pp_data = numpy.zeros((n_rows, 0), dtype=uint32)
        fp_size:  NDArray[uint32]
        fp_size = numpy.zeros((n_rows, 1), dtype=uint32)
        pp_size:  NDArray[uint32]
        pp_size = numpy.zeros((n_rows, 1), dtype=uint32)
    else:
        assert isinstance(synapse_dynamics, AbstractPlasticSynapseDynamics)
        # Blank the static data
        ff_data = [numpy.zeros(0, dtype=uint32) for _ in range(n_rows)]
        ff_size = numpy.zeros((n_rows, 1), dtype=uint32)

        # Get the plastic data
        fp_data, pp_data, fp_size, pp_size = \
            synapse_dynamics.get_plastic_synaptic_data(
                connections, row_indices, n_rows, n_synapse_types,
                max_row_n_synapses, max_atoms_per_core)

    # Add some padding
    row_lengths = [
        pp_data[i].size + fp_data[i].size + ff_data[i].size
        for i in range(n_rows)]
    padding = [
        numpy.zeros(max_row_n_words - row_length, dtype=uint32)
        for row_length in row_lengths]

    # Join the bits into rows
    rows = [numpy.concatenate(items) for items in zip(
        pp_size, pp_data, ff_size, fp_size, ff_data, fp_data, padding)]
    row_data = numpy.concatenate(rows)

    # Return the data
    return row_data


def convert_to_connections(
        synapse_info: SynapseInformation, post_vertex_slice: Slice,
        n_pre_atoms: int, max_row_length: int, n_synapse_types: int,
        weight_scales: WeightScales, data: Union[bytes, NDArray, None],
        delayed: bool, post_vertex_max_delay_ticks: int,
        max_atoms_per_core: int) -> ConnectionsArray:
    """
    Read the synapses for a given projection synapse information
    object out of the given data and convert to connection data

    :param synapse_info:
        The synapse information of the synapses
    :param n_pre_atoms: The number of atoms in the pre-vertex
    :param post_vertex_slice:
        The slice of the target neurons of the synapses in the data
    :param max_row_length:
        The length of each row in the data
    :param n_synapse_types:
        The number of synapse types in total
    :param weight_scales:
        The weight scaling of each synapse type
    :param data:
        The raw data containing the synapses
    :param delayed:
        True if the data should be considered delayed
    :param post_vertex_max_delay_ticks:
        The maximum delayed ticks supported from post vertex
    :param max_atoms_per_core:
        The maximum number of atoms on a core
    :return: The connections read from the data; the dtype is
        :py:const:`~.NUMPY_CONNECTORS_DTYPE`
    """
    # If there is no data, return nothing
    if data is None or len(data) == 0:
        return numpy.zeros(0, dtype=NUMPY_CONNECTORS_DTYPE)

    # Translate the data into rows
    row_data = numpy.frombuffer(data, dtype="<u4").reshape(
        -1, (max_row_length + _N_HEADER_WORDS))

    dynamics = synapse_info.synapse_dynamics
    if isinstance(dynamics, AbstractStaticSynapseDynamics):
        # Read static data
        connections = _read_static_data(
            dynamics, n_pre_atoms, n_synapse_types, row_data, delayed,
            post_vertex_max_delay_ticks, max_atoms_per_core)
    elif isinstance(dynamics, AbstractPlasticSynapseDynamics):
        # Read plastic data
        connections = _read_plastic_data(
            dynamics, n_pre_atoms, n_synapse_types, row_data, delayed,
            post_vertex_max_delay_ticks, max_atoms_per_core)
    else:
        raise TypeError(f"{dynamics=} has unexpected type {type(dynamics)}")

    # There might still be no connections if the row was all padding
    if not connections.size:
        return numpy.zeros(0, dtype=NUMPY_CONNECTORS_DTYPE)

    # Convert 0 delays to max delays
    connections["delay"][connections["delay"] == 0] = (
        post_vertex_max_delay_ticks)

    connections = __convert_sources_and_targets(
        connections, synapse_info.pre_vertex, post_vertex_slice)

    # Return the connections after appropriate scaling
    return _rescale_connections(connections, weight_scales, synapse_info)


def read_all_synapses(
        data: NDArray[uint32], delayed_data: NDArray[uint32],
        synapse_info: SynapseInformation, n_synapse_types: int,
        weight_scales: WeightScales, post_vertex_slice: Slice,
        n_pre_atoms: int, post_vertex_max_delay_ticks: int,
        max_row_info: MaxRowInfo, max_atoms_per_core: int
        ) -> ConnectionsArray:
    """
    Read the synapses for a given projection synapse information
    object out of the given delayed and undelayed data.

    :param data:
        The raw data containing the undelayed synapses
    :param delayed_data:
        The raw data containing the delayed synapses
    :param synapse_info:
        The synapse info that generated the synapses
    :param n_synapse_types:
        The total number of synapse types available
    :param weight_scales:
        A weight scale for each synapse type
    :param n_pre_atoms: The number of atoms in the pre-vertex
    :param post_vertex_slice:
        The slice of the post-vertex to read the synapses for
    :param post_vertex_max_delay_ticks:
            max delayed ticks supported from post vertex
    :param max_row_info:
        The maximum information for each of the rows
    :param max_atoms_per_core:
        The maximum number of atoms on a core
    :return: The connections read from the data; the dtype is
        :py:const:`~.NUMPY_CONNECTORS_DTYPE`
    """
    connections: List[ConnectionsArray] = []
    max_row_length = max_row_info.undelayed_max_words
    delayed_max_row_length = max_row_info.delayed_max_words
    connections.append(convert_to_connections(
        synapse_info, post_vertex_slice, n_pre_atoms, max_row_length,
        n_synapse_types, weight_scales, data, False,
        post_vertex_max_delay_ticks, max_atoms_per_core))
    connections.append(convert_to_connections(
        synapse_info, post_vertex_slice, n_pre_atoms,
        delayed_max_row_length, n_synapse_types, weight_scales, delayed_data,
        True, post_vertex_max_delay_ticks, max_atoms_per_core))

    # Join the connections into a single list and return it
    return numpy.concatenate(connections)


def _parse_static_data(
        row_data: _RowData,
        dynamics: AbstractStaticSynapseDynamics) -> Tuple[
            NDArray[numpy.integer], List[_RowData]]:
    """
    Parse static synaptic data.

    :param row_data: The raw row data
    :param dynamics:
        The synapse dynamics that can decode the rows
    :return: A tuple of the recorded length of each row and the row data
        organised into rows
    """
    n_rows = row_data.shape[0]
    ff_size = row_data[:, 1]
    ff_words = dynamics.get_n_static_words_per_row(ff_size)
    ff_start = _N_HEADER_WORDS
    ff_end = ff_start + ff_words
    return (
        ff_size,
        [row_data[row, ff_start:ff_end[row]] for row in range(n_rows)])


def _read_static_data(
        dynamics: AbstractStaticSynapseDynamics, n_pre_atoms: int,
        n_synapse_types: int, row_data: _RowData, delayed: bool,
        post_vertex_max_delay_ticks: int,
        max_atoms_per_core: int) -> ConnectionsArray:
    """
    Read static data from row data.

    :param dynamics:
        The synapse dynamics that generated the data
    :param n_pre_atoms: The number of atoms on the pre-vertex
    :param n_synapse_types:
        The number of synapse types available
    :param row_data:
        The raw row data to read
    :param delayed: True if data should be considered delayed
    :param post_vertex_max_delay_ticks: post vertex delay maximum
    :param max_atoms_per_core: The maximum number of atoms on a core
    :return: the connections read with dtype
        :py:const:`~.NUMPY_CONNECTORS_DTYPE`
    """
    if row_data is None or not row_data.size:
        return numpy.zeros(0, dtype=NUMPY_CONNECTORS_DTYPE)
    ff_size, ff_data = _parse_static_data(row_data, dynamics)
    connections = dynamics.read_static_synaptic_data(
        n_synapse_types, ff_size, ff_data, max_atoms_per_core)
    if delayed:
        n_synapses = dynamics.get_n_synapses_in_rows(ff_size)
        connections = _convert_delayed_data(
            n_synapses, n_pre_atoms, connections,
            post_vertex_max_delay_ticks)
    return connections


def _parse_plastic_data(
        row_data: _RowData,
        dynamics: AbstractPlasticSynapseDynamics) -> Tuple[
            NDArray[uint32], List[_RowData], NDArray[uint32], List[_RowData]]:
    """
    Parse plastic synapses from raw row data.

    :param row_data: The raw data to parse
    :param dynamics:
        The dynamics that generated the data
    :return: A tuple of the recorded length of the plastic-plastic data in
        each row; the plastic-plastic data organised into rows; the
        recorded length of the static-plastic data in each row; and the
        static-plastic data organised into rows
    """
    n_rows = row_data.shape[0]
    pp_size = row_data[:, 0]
    pp_words = dynamics.get_n_plastic_plastic_words_per_row(pp_size)
    fp_size = row_data[numpy.arange(n_rows), pp_words + 2]
    fp_words = dynamics.get_n_fixed_plastic_words_per_row(fp_size)
    fp_start = pp_size + _N_HEADER_WORDS
    fp_end = fp_start + fp_words
    row_ids = range(n_rows)
    return (
        pp_size,
        [row_data[row, 1:pp_words[row] + 1] for row in row_ids],
        fp_size,
        [row_data[row, fp_start[row]:fp_end[row]] for row in row_ids])


def _read_plastic_data(
        dynamics: AbstractPlasticSynapseDynamics, n_pre_atoms: int,
        n_synapse_types: int, row_data: Optional[_RowData], delayed: bool,
        post_vertex_max_delay_ticks: int,
        max_atoms_per_core: int) -> ConnectionsArray:
    """
    Read plastic data from raw data.

    :param dynamics:
        The synapse dynamics that generated the data
    :param n_pre_atoms: The number of atoms in the pre-vertex
    :param n_synapse_types:
        The number of synapse types available
    :param row_data:
        The raw row data to read
    :param delayed: True if data should be considered delayed
    :param post_vertex_max_delay_ticks: post vertex delay maximum
    :param max_atoms_per_core: The maximum number of atoms on a core
    :return: the connections read with dtype
        :py:const:`~.NUMPY_CONNECTORS_DTYPE`
    """
    if row_data is None or not row_data.size:
        return numpy.zeros(0, dtype=NUMPY_CONNECTORS_DTYPE)
    pp_size, pp_data, fp_size, fp_data = _parse_plastic_data(
        row_data, dynamics)
    connections = dynamics.read_plastic_synaptic_data(
        n_synapse_types, pp_size, pp_data, fp_size, fp_data,
        max_atoms_per_core)

    if delayed:
        n_synapses = dynamics.get_n_synapses_in_rows(pp_size, fp_size)
        connections = _convert_delayed_data(
            n_synapses, n_pre_atoms, connections,
            post_vertex_max_delay_ticks)
    return connections


def _rescale_connections(
        connections: ConnectionsArray, weight_scales: WeightScales,
        synapse_info: SynapseInformation) -> ConnectionsArray:
    """
    Scale the connection data into machine values.

    :param connections: The connections to be rescaled
    :param weight_scales: The weight scale of each synapse type
    :param synapse_info: The synapse information of the connections
    """
    # Return the delays values to milliseconds
    connections["delay"] /= SpynnakerDataView.get_simulation_time_step_per_ms()
    # Undo the weight scaling
    connections["weight"] /= weight_scales[synapse_info.synapse_type]
    return connections


def _convert_delayed_data(
        n_synapses: NDArray[integer], n_pre_atoms: int,
        delayed_connections: ConnectionsArray,
        post_vertex_max_delay_ticks: int) -> ConnectionsArray:
    """
    Take the delayed_connections and convert the source ids and delay
    values back to global values.

    :param n_synapses: The number of synapses in each row
    :param n_pre_atoms: number of atoms in the pre-vertex
    :param delayed_connections:
        The connections to convert of dtype
        :py:const:`~.NUMPY_CONNECTORS_DTYPE`
    :param post_vertex_max_delay_ticks: post vertex delay maximum
    :return: The converted connection with the same dtype
    """
    # Work out the delay stage of each row; rows are the all the rows
    # from the first delay stage, then all from the second stage and so on
    synapse_ids = range(len(n_synapses))
    row_stage = numpy.array([
        i // n_pre_atoms
        for i in synapse_ids], dtype=uint32)
    # Work out the delay for each stage
    row_min_delay = (row_stage + 1) * post_vertex_max_delay_ticks
    # Repeat the delay for all connections in the same row
    connection_min_delay = numpy.concatenate([
        numpy.repeat(row_min_delay[i], n_synapses[i])
        for i in synapse_ids])
    # Repeat the "extra" source id for all connections in the same row;
    # this converts the row id back to a source neuron id
    connection_source_extra = numpy.concatenate([
        numpy.repeat(
            row_stage[i] * uint32(n_pre_atoms),
            n_synapses[i])
        for i in synapse_ids])
    # Do the conversions
    delayed_connections["source"] -= connection_source_extra
    delayed_connections["delay"] += connection_min_delay
    return delayed_connections


def __convert_sources_and_targets(
        connections: ConnectionsArray, pre_vertex: ApplicationVertex,
        post_vertex_slice: Slice) -> ConnectionsArray:
    connections["source"] = pre_vertex.get_raster_ordered_indices(
        connections["source"])
    connections["target"] = post_vertex_slice.get_raster_indices(
        connections["target"])
    return connections



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_front_end_common.interface.ds import DataType
from spynnaker.pyNN.protocols import MunichIoSpiNNakerLinkProtocol
from spynnaker.pyNN.external_devices_models.push_bot import (
    AbstractPushBotOutputDevice)


class PushBotLaser(AbstractPushBotOutputDevice):
    """
    The properties of the laser device that may be set.
    """

    #: The total period for the laser
    LASER_TOTAL_PERIOD = (
        0,
        MunichIoSpiNNakerLinkProtocol.push_bot_laser_config_total_period_key,
        0, DataType.S1615.max, 20
    )

    #: The active period for the laser (no larger than the total period)
    LASER_ACTIVE_TIME = (
        1, MunichIoSpiNNakerLinkProtocol.push_bot_laser_config_active_time_key,
        0, DataType.S1615.max, 20
    )

    #: The frequency of the laser
    LASER_FREQUENCY = (
        2, MunichIoSpiNNakerLinkProtocol.push_bot_laser_set_frequency_key,
        0, DataType.S1615.max, 20
    )



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import List, Sequence, Tuple, TYPE_CHECKING

from numpy import floating
from numpy.typing import NDArray

from spinn_utilities.overrides import overrides
from spinn_utilities.abstract_base import abstractmethod

from pacman.model.graphs.common import Slice
from pacman.model.graphs.machine import MachineVertex
from pacman.model.placements import Placement

from spinn_front_end_common.utilities.helpful_functions import (
    locate_memory_region_for_placement)
from spinn_front_end_common.abstract_models import (
    AbstractSupportsBitFieldRoutingCompression)
from spinn_front_end_common.interface.ds import DataSpecificationBase
from spinn_front_end_common.interface.buffer_management.buffer_models import (
    AbstractReceiveRegionsToHost)

from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    AbstractSynapseDynamicsStructural, AbstractSDRAMSynapseDynamics)
from spynnaker.pyNN.utilities.utility_calls import get_n_bits
from spynnaker.pyNN.models.abstract_models import (
    AbstractSynapseExpandable, HasSynapses)

from .population_vertex import PopulationVertex
from .population_machine_synapses_provenance import (
    PopulationMachineSynapsesProvenance)
from .synaptic_matrices import SynapseRegions, SynapseRegionReferences

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neuron.synaptic_matrices import SynapticMatrices
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)


class PopulationMachineSynapses(
        PopulationMachineSynapsesProvenance,
        AbstractSupportsBitFieldRoutingCompression,
        AbstractSynapseExpandable,
        HasSynapses, AbstractReceiveRegionsToHost, allow_derivation=True):
    """
    Mix-in for machine vertices that contain synapses.
    """

    # This MUST stay empty to allow mixing with other things with slots
    __slots__ = ()

    @property
    @abstractmethod
    def _pop_vertex(self) -> PopulationVertex:
        """
        The application vertex of the machine vertex.

        .. note::
            This is likely to be available via the MachineVertex.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    @overrides(MachineVertex.vertex_slice)
    def vertex_slice(self) -> Slice:
        # classes that implement this API also implement MachineVertex
        # pylint: disable=missing-function-docstring
        raise NotImplementedError

    @property
    @abstractmethod
    def _synaptic_matrices(self) -> SynapticMatrices:
        """
        The object holding synaptic matrices.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def _synapse_regions(self) -> SynapseRegions:
        """
        The identifiers of synaptic regions.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def _max_atoms_per_core(self) -> int:
        """
        The maximum number of atoms on any core targeted by these synapses.
        """
        raise NotImplementedError

    @abstractmethod
    def set_do_synapse_regeneration(self) -> None:
        """
        Indicates that synaptic data regeneration is required.
        """
        raise NotImplementedError

    @property
    def _synapse_references(self) -> SynapseRegionReferences:
        """
        The references to synapse regions.  Override to provide these.
        """
        return SynapseRegionReferences()

    @overrides(AbstractSupportsBitFieldRoutingCompression.
               bit_field_base_address)
    def bit_field_base_address(self, placement: Placement) -> int:
        return locate_memory_region_for_placement(
            placement=placement, region=self._synapse_regions.bitfield_filter)

    @overrides(AbstractSupportsBitFieldRoutingCompression.
               regeneratable_sdram_blocks_and_sizes)
    def regeneratable_sdram_blocks_and_sizes(
            self, placement: Placement) -> List[Tuple[int, int]]:
        synaptic_matrix_base_address = locate_memory_region_for_placement(
            placement=placement, region=self._synapse_regions.synaptic_matrix)
        return [(
            self._synaptic_matrices.host_generated_block_addr +
            synaptic_matrix_base_address,
            self._synaptic_matrices.on_chip_generated_matrix_size)]

    def _write_synapse_data_spec(
            self, spec: DataSpecificationBase,
            ring_buffer_shifts: Sequence[int],
            weight_scales: NDArray[floating], structural_sz: int) -> None:
        """
        Write the data specification for the synapse data.

        :param spec: The data specification to write to
        :param ring_buffer_shifts:
            The shifts to apply to convert ring buffer values to S1615 values
        :param weight_scales:
            The scaling to apply to weights to store them in the synapses
        :param structural_sz: The size of the structural data
        """
        # Write the synapse parameters
        self._write_synapse_parameters(spec, ring_buffer_shifts)

        # Write the synaptic matrices
        self._synaptic_matrices.generate_data()
        self._synaptic_matrices.write_synaptic_data(
            spec, self.vertex_slice, self._synapse_references)

        # Write any synapse dynamics
        synapse_dynamics = self._pop_vertex.synapse_dynamics
        synapse_dynamics_sz = self._pop_vertex.get_synapse_dynamics_size(
            self.vertex_slice.n_atoms)
        if synapse_dynamics_sz > 0:
            assert isinstance(synapse_dynamics, AbstractSDRAMSynapseDynamics)
            spec.reserve_memory_region(
                region=self._synapse_regions.synapse_dynamics,
                size=synapse_dynamics_sz, label='synapseDynamicsParams',
                reference=self._synapse_references.synapse_dynamics)
            synapse_dynamics.write_parameters(
                spec, self._synapse_regions.synapse_dynamics,
                self._pop_vertex.weight_scale, weight_scales)
        elif self._synapse_references.synapse_dynamics is not None:
            # If there is a reference for this region, we have to create it!
            spec.reserve_memory_region(
                region=self._synapse_regions.synapse_dynamics,
                size=4, label='synapseDynamicsParams',
                reference=self._synapse_references.synapse_dynamics)
        if isinstance(synapse_dynamics, AbstractSynapseDynamicsStructural):
            spec.reserve_memory_region(
                region=self._synapse_regions.structural_dynamics,
                size=structural_sz, label='synapseDynamicsStructuralParams',
                reference=self._synapse_references.structural_dynamics)
            synapse_dynamics.write_structural_parameters(
                spec, self._synapse_regions.structural_dynamics,
                weight_scales, self._pop_vertex, self.vertex_slice,
                self._synaptic_matrices)
        elif self._synapse_references.structural_dynamics is not None:
            # If there is a reference for this region, we have to create it!
            spec.reserve_memory_region(
                region=self._synapse_regions.structural_dynamics,
                size=4, label='synapseDynamicsStructuralParams',
                reference=self._synapse_references.structural_dynamics)

    def _write_synapse_parameters(
            self, spec: DataSpecificationBase,
            ring_buffer_shifts: Sequence[int]) -> None:
        """
        Write the synapse parameters data region.

        :param spec: The data specification to write to
        :param ring_buffer_shifts:
            The shifts to apply to convert ring buffer values to S1615 values
        """
        # Reserve space
        spec.reserve_memory_region(
            region=self._synapse_regions.synapse_params,
            size=self._pop_vertex.get_synapse_params_size(),
            label='SynapseParams',
            reference=self._synapse_references.synapse_params)

        # Get values
        n_neurons = self._max_atoms_per_core
        # We only count neuron synapse types here, as this is related to
        # the ring buffers
        n_synapse_types = self._pop_vertex.neuron_impl.get_n_synapse_types()
        max_delay = self._pop_vertex.splitter.max_support_delay()

        # Write synapse parameters
        spec.switch_write_focus(self._synapse_regions.synapse_params)
        spec.write_value(n_neurons)
        spec.write_value(n_synapse_types)
        spec.write_value(get_n_bits(n_neurons))
        spec.write_value(get_n_bits(n_synapse_types))
        spec.write_value(get_n_bits(max_delay))
        spec.write_value(int(self._pop_vertex.drop_late_spikes))
        spec.write_value(self._pop_vertex.incoming_spike_buffer_size)
        spec.write_array(ring_buffer_shifts)

    @overrides(AbstractSynapseExpandable.gen_on_machine)
    def gen_on_machine(self) -> bool:
        return self._synaptic_matrices.gen_on_machine

    @overrides(AbstractSynapseExpandable.read_generated_connection_holders)
    def read_generated_connection_holders(self, placement: Placement) -> None:
        self._synaptic_matrices.read_generated_connection_holders(placement)

    @property
    @overrides(AbstractSynapseExpandable.connection_generator_region)
    def connection_generator_region(self) -> int:
        return self._synapse_regions.connection_builder

    @overrides(HasSynapses.get_connections_from_machine)
    def get_connections_from_machine(
            self, placement: Placement, app_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> Sequence[NDArray]:
        return self._synaptic_matrices.get_connections_from_machine(
            placement, app_edge, synapse_info)

    @property
    @overrides(AbstractSynapseExpandable.max_gen_data)
    def max_gen_data(self) -> int:
        return self._synaptic_matrices.max_gen_data

    @property
    @overrides(AbstractSynapseExpandable.bit_field_size)
    def bit_field_size(self) -> int:
        return self._synaptic_matrices.bit_field_size

    @overrides(AbstractReceiveRegionsToHost.get_download_regions)
    def get_download_regions(
            self, placement: Placement) -> Sequence[Tuple[int, int, int]]:

        # Make sure not to overwrite the recording regions
        return self._synaptic_matrices.get_download_regions(placement)



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import math
from typing import Optional, Union
from spinn_utilities.config_holder import get_config_int
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinn_front_end_common.data.fec_data_writer import FecDataWriter
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from .spynnaker_data_view import SpynnakerDataView, _SpynnakerDataModel

logger = FormatAdapter(logging.getLogger(__name__))
# pylint: disable=protected-access


class SpynnakerDataWriter(FecDataWriter, SpynnakerDataView):
    """
    See :py:class:`~spinn_utilities.data.utils_data_writer.UtilsDataWriter`.

    This class is designed to only be used directly by
    :py:class:`spinn_front_end_common.interface.abstract_spinnaker_base.AbstractSpinnakerBase`
    and its subclasses and within the PyNN repositories unit tests.
    """

    __spy_data = _SpynnakerDataModel()

    @overrides(FecDataWriter._setup)
    def _setup(self) -> None:
        FecDataWriter._setup(self)
        self.__spy_data._clear()

    @overrides(FecDataWriter._mock)
    def _mock(self) -> None:
        FecDataWriter._mock(self)
        self._set_min_delay(1)

    @overrides(FecDataWriter._hard_reset)
    def _hard_reset(self) -> None:
        FecDataWriter._hard_reset(self)
        self.__spy_data._hard_reset()

    @overrides(FecDataWriter._soft_reset)
    def _soft_reset(self) -> None:
        FecDataWriter._soft_reset(self)
        self.__spy_data._soft_reset()

    def set_up_timings_and_delay(
            self, simulation_time_step_us: Optional[int],
            time_scale_factor: Optional[float],
            min_delay: Optional[Union[int, float]]) -> None:
        """
        Setup the timings and delays for the simulation

        :param simulation_time_step_us:
            An explicitly specified time step for the simulation in
            microseconds.
            If `None`, the value is read from the configuration
        :param time_scale_factor:
            An explicitly specified time scale factor for the simulation.
            If `None`, the value is read from the configuration
        :param min_delay:
            new value or `None` to say use simulation_time_step_ms
        """
        try:
            # If there isn't a time scale factor given, try to work it out
            # on the basis that we can do 0.1ms steps in real time
            if time_scale_factor is None:
                # Need to do this early to ensure we have a value
                if simulation_time_step_us is None:
                    simulation_time_step_us = get_config_int(
                        "Machine", "simulation_time_step")
                default_time_scale_factor = max(
                    1.0,
                    math.ceil(simulation_time_step_us / 10000.0))
                self.set_up_timings(
                    simulation_time_step_us, time_scale_factor,
                    default_time_scale_factor)
                if self.get_time_scale_factor() > 1.0:
                    logger.warning(
                        "A timestep was entered that has forced SpiNNaker to "
                        "automatically slow the simulation down from real "
                        f"time by a factor of {time_scale_factor}.")
            else:
                self.set_up_timings(simulation_time_step_us, time_scale_factor)
            self._set_min_delay(min_delay)
        except ConfigurationException:
            self.__spy_data._min_delay = None
            raise

    def _set_min_delay(self, min_delay: Optional[Union[int, float]]) -> None:
        """
        Sets a min delay or accepts `None` to use simulation_time_step_ms.

        :param min_delay:
            new value or `None` to say use simulation_time_step_ms
        """
        if min_delay is None:
            min_delay = self.get_simulation_time_step_ms()

        if not isinstance(min_delay, (int, float)):
            raise TypeError("min_delay should be an float (or int)")

        if min_delay < self.get_simulation_time_step_ms():
            raise ConfigurationException(
                f'invalid min_delay: {min_delay} '
                f'must at least simulation time step in microseconds: '
                f'{self.get_simulation_time_step_ms()}')

        raw = min_delay / self.get_simulation_time_step_ms()
        rounded = round(raw)
        if abs(rounded - raw) > 0.00001:
            raise ConfigurationException(
                f'invalid min_delay {min_delay} '
                f'must be a multiple of simulation time step in microseconds '
                f'{self.get_simulation_time_step_ms()}')

        self.__spy_data._min_delay = min_delay

    def _get_id_counter(self) -> int:
        """
        Testing method likely to change without notice!
        """
        return self.__spy_data._id_counter



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, Optional
from spinn_utilities.overrides import overrides
from pacman.model.graphs.application import (
    ApplicationSpiNNakerLinkVertex)
from pacman.model.graphs.application import ApplicationVertex
from pacman.model.graphs.application.abstract import (
    AbstractOneAppOneMachineVertex)
from spinn_front_end_common.abstract_models import (
    AbstractVertexWithEdgeToDependentVertices)
from spynnaker.pyNN.models.defaults import AbstractProvidesDefaults
from spynnaker.pyNN.models.common import PopulationApplicationVertex
from .machine_munich_motor_device import MachineMunichMotorDevice


class _MunichMotorDevice(ApplicationSpiNNakerLinkVertex):
    __slots__ = ()

    def __init__(self, spinnaker_link_id: int,
                 board_address: Optional[str] = None):
        """
        :param spinnaker_link_id:
            The index of the spinnaker link to which the device is connected
        :param board_address:
            The optional IP address of the board to which the device is
            connected e.g. in a multi-board system
        """
        super().__init__(
            n_atoms=6, spinnaker_link_id=spinnaker_link_id,
            label="External Munich Motor", board_address=board_address)


class MunichMotorDevice(
        AbstractOneAppOneMachineVertex,
        AbstractVertexWithEdgeToDependentVertices,
        PopulationApplicationVertex,
        AbstractProvidesDefaults):
    """
    An Omnibot motor control device. This has a real vertex and an
    external device vertex.
    """

    __slots__ = ("__dependent_vertices", )

    def __init__(
            self, spinnaker_link_id: int, board_address: Optional[str] = None,
            speed: int = 30, sample_time: int = 4096, update_time: int = 512,
            delay_time: int = 5, delta_threshold: int = 23,
            continue_if_not_different: bool = True,
            label: Optional[str] = None):
        """
        :param spinnaker_link_id:
            The SpiNNaker link to which the motor is connected
        :param board_address:
        :param speed:
        :param sample_time:
        :param update_time:
        :param delay_time:
        :param delta_threshold:
        :param continue_if_not_different:
        :param label: The optional name of the vertex.
        """
        m_vertex = MachineMunichMotorDevice(
            speed, sample_time, update_time, delay_time, delta_threshold,
            continue_if_not_different, label, app_vertex=self)
        super().__init__(
            m_vertex, label, MachineMunichMotorDevice._N_ATOMS)
        self.__dependent_vertices = [
            _MunichMotorDevice(spinnaker_link_id, board_address)]

    @overrides(AbstractVertexWithEdgeToDependentVertices.dependent_vertices)
    def dependent_vertices(self) -> Iterable[ApplicationVertex]:
        return self.__dependent_vertices

    @overrides(AbstractVertexWithEdgeToDependentVertices.
               edge_partition_identifiers_for_dependent_vertex)
    def edge_partition_identifiers_for_dependent_vertex(
            self, vertex: ApplicationVertex) -> Iterable[str]:
        yield self.machine_vertex.MOTOR_PARTITION_ID



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .abstract_ethernet_controller import AbstractEthernetController
from .abstract_ethernet_sensor import AbstractEthernetSensor
from .abstract_ethernet_translator import AbstractEthernetTranslator
from .abstract_multicast_controllable_device import (
    AbstractMulticastControllableDevice, SendType)
from .arbitrary_fpga_device import ArbitraryFPGADevice
from .external_device_lif_control import ExternalDeviceLifControl
from .external_spinnaker_link_fpga_retina_device import (
    ExternalFPGARetinaDevice)
from .machine_munich_motor_device import MachineMunichMotorDevice
from .munich_spinnaker_link_motor_device import MunichMotorDevice
from .munich_spinnaker_link_retina_device import MunichRetinaDevice
from .threshold_type_multicast_device_control import (
    ThresholdTypeMulticastDeviceControl)
from .spif_retina_device import SPIFRetinaDevice
from .icub_retina_device import ICUBRetinaDevice
from .spif_output_device import SPIFOutputDevice
from .spif_input_device import SPIFInputDevice

__all__ = ["AbstractEthernetController", "AbstractEthernetSensor",
           "AbstractEthernetTranslator", "ArbitraryFPGADevice",
           "AbstractMulticastControllableDevice", "ExternalDeviceLifControl",
           "ExternalFPGARetinaDevice", "MachineMunichMotorDevice",
           "MunichMotorDevice", "MunichRetinaDevice", "SendType",
           "ThresholdTypeMulticastDeviceControl", "SPIFRetinaDevice",
           "ICUBRetinaDevice", "SPIFOutputDevice", "SPIFInputDevice"]



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.require_subclass import require_subclass
from pacman.exceptions import PacmanConfigurationException
from pacman.model.graphs.application import ApplicationVertex
if TYPE_CHECKING:
    from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
        ConnectionsArray)
    from spynnaker.pyNN.models.neuron.synapse_dynamics import (
        AbstractSynapseDynamics)
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)
    from spynnaker.pyNN.extra_algorithms.splitter_components import (
        AbstractSpynnakerSplitterDelay)


@require_subclass(ApplicationVertex)
class AbstractAcceptsIncomingSynapses(object, metaclass=AbstractBase):
    """
    Indicates an application vertex that can be a post-vertex in a PyNN
    projection.

    .. note::
        See :py:meth:`verify_splitter`
    """
    __slots__ = ()

    @abstractmethod
    def get_synapse_id_by_target(self, target: str) -> Optional[int]:
        """
        :param target: The name of the synapse
        :returns: The ID of the synapse with this name
        """
        raise NotImplementedError

    @abstractmethod
    def set_synapse_dynamics(
            self, synapse_dynamics: AbstractSynapseDynamics) -> None:
        """
        Set the synapse dynamics of this vertex.

        :param synapse_dynamics:
        """
        raise NotImplementedError

    @abstractmethod
    def get_connections_from_machine(
            self, app_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> ConnectionsArray:
        """
        :param app_edge:
            The edge for which the data is being read
        :param synapse_info:
            The specific projection within the edge
        :returns: The connections from the machine post-run.
        """
        raise NotImplementedError

    @abstractmethod
    def clear_connection_cache(self) -> None:
        """
        Clear the connection data stored in the vertex so far.
        """
        raise NotImplementedError

    def verify_splitter(
            self, splitter: AbstractSpynnakerSplitterDelay) -> None:
        """
        Check that the splitter implements the API(s) expected by the
        SynapticMatrices

        Any Vertex that implements this API should override
        `ApplicationVertex.splitter` method to also call this function.

        :param splitter: the splitter
        :raises PacmanConfigurationException: if the splitter is not an
            instance of AbstractSpynnakerSplitterDelay
        """
        # Delayed import to avoid circular dependency
        # pylint: disable=import-outside-toplevel
        from spynnaker.pyNN.extra_algorithms.splitter_components import (
            AbstractSpynnakerSplitterDelay as DelaySplitter)
        if not isinstance(splitter, DelaySplitter):
            raise PacmanConfigurationException(
                "The splitter needs to be an instance of "
                "----------------AbstractSpynnakerSplitterDelay")



# Copyright (c) 2023 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from math import ceil, log2, floor
from collections import namedtuple, defaultdict
from typing import Dict, List, Tuple, TYPE_CHECKING, Union

from pacman.model.graphs.application import (
    ApplicationVertex, ApplicationVirtualVertex)
from pacman.model.graphs.common.slice import Slice
from pacman.model.graphs.common.mdslice import MDSlice
from pacman.model.routing_info import AppVertexRoutingInfo

from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from spynnaker.pyNN.data.spynnaker_data_view import SpynnakerDataView
from spynnaker.pyNN.models.abstract_models import ColouredApplicationVertex
from spynnaker.pyNN.utilities.utility_calls import get_n_bits

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neuron import PopulationVertex
    from spynnaker.pyNN.models.projection import Projection

#: The number of bits in a short value
BITS_PER_SHORT = 16

#: The number of bits in a byte value
BITS_PER_BYTE = 8

#: The number of bits to represent n_colour_bits
N_COLOUR_BITS_BITS = 3

#: Key info size in bytes
KEY_INFO_SIZE = 4 * BYTES_PER_WORD

#: A source
Source = namedtuple(
    "Source", ["projection", "local_delay", "delay_stage"])


def get_div_const(value: int) -> int:
    """ Get the values used to perform fast division by an integer constant

    :param value: The value to be divided by
    :return: The values required encoded as fields of a 32-bit integer
    """
    log_val = int(ceil(log2(value)))
    log_m_val = ((2 ** log_val) - value) / value
    m = int(floor((2 ** BITS_PER_SHORT) * log_m_val) + 1)
    sh1 = min(log_val, 1)
    sh2 = max(log_val - 1, 0)
    return ((sh2 << (BITS_PER_SHORT + BITS_PER_BYTE))
            + (sh1 << BITS_PER_SHORT) + m)


def get_delay_for_source(incoming: "Projection") -> Tuple[
        ColouredApplicationVertex, int, int, str]:
    """ Get the vertex which will send data from a given source projection,
        along with the delay stage and locally-handled delay value

    :param incoming: The incoming projection to get the delay from
    :return: The vertex, the local delay, the delay stage, the partition id
    """
    # pylint: disable=protected-access
    app_edge = incoming._projection_edge
    s_info = incoming._synapse_information
    delay = s_info.synapse_dynamics.delay
    if not isinstance(delay, float):
        raise NotImplementedError("Only float delay supported for now")
    steps = delay * SpynnakerDataView.get_simulation_time_step_per_ms()
    max_delay = app_edge.post_vertex.splitter.max_support_delay()
    local_delay = int(steps % max_delay)
    delay_stage: int = 0
    pre_vertex: ColouredApplicationVertex = app_edge.pre_vertex
    if steps > max_delay:
        delay_stage = int(steps // max_delay) - 1
        delay_edge = app_edge.delay_edge
        assert delay_edge is not None
        pre_vertex = delay_edge.pre_vertex
    return pre_vertex, local_delay, delay_stage, s_info.partition_id


def get_rinfo_for_spike_source(
        pre_vertex: ApplicationVertex,
        partition_id: str) -> Tuple[AppVertexRoutingInfo, int, int]:
    """
    Get the routing information for the source of a projection in the
    given partition.

    :param pre_vertex: The source of incoming data
    :param partition_id: The partition ID to get the routing info from
    :return: Routing information, core mask, core mask shift
    """
    routing_info = SpynnakerDataView.get_routing_infos()

    # Find the routing information
    r_info = routing_info.get_info_from(
            pre_vertex, partition_id)
    assert isinstance(r_info, AppVertexRoutingInfo)

    n_cores = len(r_info.vertex.splitter.get_out_going_vertices(partition_id))

    # If there is 1 core, we don't use the core mask
    # If there is a virtual vertex, these also don't use core masks
    if n_cores == 1 or isinstance(pre_vertex, ApplicationVirtualVertex):
        return r_info, 0, 0

    mask_shift = r_info.n_bits_atoms
    core_mask = (2 ** get_n_bits(n_cores)) - 1
    return r_info, core_mask, mask_shift


def get_sources_for_target(app_vertex: "PopulationVertex") -> Dict[
        Tuple[ColouredApplicationVertex, str], List[Source]]:
    """
    Get all the application vertex sources that will hit the given application
    vertex.

    :param app_vertex: The vertex being targeted
    :return:
        A dict of source ApplicationVertex to list of source information
    """
    sources = defaultdict(list)
    for incoming in app_vertex.incoming_projections:
        pre_vertex, local_delay, delay_stage, part_id = get_delay_for_source(
            incoming)
        source = Source(incoming, local_delay, delay_stage)
        sources[pre_vertex, part_id].append(source)
    return sources


def get_first_and_last_slice(pre_vertex: ApplicationVertex) -> \
        Union[Tuple[Slice, Slice], Tuple[MDSlice, MDSlice]]:
    """
    :param pre_vertex: The source vertex
    :returns: The first and last slice of an application vertex.
    """
    if isinstance(pre_vertex, ApplicationVirtualVertex):
        if len(pre_vertex.atoms_shape) == 1:
            full_slice = Slice(0, pre_vertex.n_atoms - 1)
            return full_slice, full_slice
        atoms_shape = pre_vertex.atoms_shape
        full_slice = MDSlice(
            0, pre_vertex.n_atoms - 1, atoms_shape,
            tuple(0 for _ in atoms_shape), atoms_shape)
        return full_slice, full_slice
    m_vertices = list(pre_vertex.machine_vertices)
    return m_vertices[0].vertex_slice, m_vertices[-1].vertex_slice



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import IntEnum
import ctypes
from typing import List, Optional, Sequence

from numpy import floating
from numpy.typing import NDArray

from spinn_utilities.overrides import overrides

from pacman.model.graphs.machine import (
    MachineVertex, SDRAMMachineEdge, SourceSegmentedSDRAMMachinePartition)
from pacman.model.graphs.common import Slice
from pacman.model.resources import AbstractSDRAM
from pacman.model.placements import Placement

from spinn_front_end_common.abstract_models import (
    AbstractGeneratesDataSpecification, AbstractRewritesDataSpecification)
from spinn_front_end_common.interface.provenance import ProvenanceWriter
from spinn_front_end_common.interface.ds import (
    DataSpecificationGenerator, DataSpecificationReloader)
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from spynnaker.pyNN.exceptions import SynapticConfigurationException
from spynnaker.pyNN.models.abstract_models import (
    ReceivesSynapticInputsOverSDRAM, SendsSynapticInputsOverSDRAM)
from spynnaker.pyNN.models.neuron.neuron_data import NeuronData
from spynnaker.pyNN.utilities.utility_calls import get_n_bits

from .population_machine_common import CommonRegions, PopulationMachineCommon
from .population_machine_neurons import (
    NeuronRegions, PopulationMachineNeurons, NeuronProvenance)
from .population_vertex import PopulationVertex

# Size of SDRAM params = 1 word for address + 1 word for size
# + 1 word for n_neurons + 1 word for n_synapse_types
# + 1 word for number of synapse vertices
# + 1 word for number of neuron bits needed
SDRAM_PARAMS_SIZE = 6 * BYTES_PER_WORD


class NeuronMainProvenance(ctypes.LittleEndianStructure):
    """
    Provenance items from synapse processing.
    """
    _fields_ = [
        # the maximum number of times the timer tick didn't complete in time
        ("n_timer_overruns", ctypes.c_uint32),
    ]

    N_ITEMS = len(_fields_)


class PopulationNeuronsMachineVertex(
        PopulationMachineCommon,
        PopulationMachineNeurons,
        AbstractGeneratesDataSpecification,
        AbstractRewritesDataSpecification,
        ReceivesSynapticInputsOverSDRAM):
    """
    A machine vertex for the Neurons of PyNN Populations.
    """

    __slots__ = (
        "__key",
        "__sdram_partition",
        "__ring_buffer_shifts",
        "__weight_scales",
        "__slice_index",
        "__neuron_data",
        "__max_atoms_per_core",
        "__regenerate_data")

    class REGIONS(IntEnum):
        """
        Regions for populations.
        """
        SYSTEM = 0
        CORE_PARAMS = 1
        PROVENANCE_DATA = 2
        PROFILING = 3
        RECORDING = 4
        NEURON_PARAMS = 5
        CURRENT_SOURCE_PARAMS = 6
        NEURON_RECORDING = 7
        SDRAM_EDGE_PARAMS = 8
        NEURON_BUILDER = 9
        INITIAL_VALUES = 10

    #: Regions for this vertex used by common parts
    COMMON_REGIONS = CommonRegions(
        REGIONS.SYSTEM,
        REGIONS.PROVENANCE_DATA,
        REGIONS.PROFILING,
        REGIONS.RECORDING)

    #: Regions for this vertex used by neuron parts
    NEURON_REGIONS = NeuronRegions(
        REGIONS.CORE_PARAMS,
        REGIONS.NEURON_PARAMS,
        REGIONS.CURRENT_SOURCE_PARAMS,
        REGIONS.NEURON_RECORDING,
        REGIONS.NEURON_BUILDER,
        REGIONS.INITIAL_VALUES)

    _PROFILE_TAG_LABELS = {
        0: "TIMER_NEURONS"}

    def __init__(
            self, sdram: AbstractSDRAM, label: str,
            app_vertex: PopulationVertex, vertex_slice: Slice,
            slice_index: int, ring_buffer_shifts: Sequence[int],
            weight_scales: NDArray[floating],
            neuron_data: NeuronData, max_atoms_per_core: int):
        """
        :param sdram: The SDRAM used by the vertex
        :param label: The label of the vertex
        :param app_vertex: The associated application vertex
        :param vertex_slice: The slice of the population that this implements
        :param slice_index:
            The index of the slice in the ordered list of slices
        :param ring_buffer_shifts:
            The shifts to apply to convert ring buffer values to S1615 values
        :param weight_scales:
            The scaling to apply to weights to store them in the synapses
        :param neuron_data: The handler of neuron data
        :param max_atoms_per_core: The maximum number of atoms per core
        """
        super().__init__(
            label, app_vertex, vertex_slice, sdram, self.COMMON_REGIONS,
            NeuronProvenance.N_ITEMS + NeuronMainProvenance.N_ITEMS,
            self._PROFILE_TAG_LABELS, app_vertex.neuron_core_binary_file_name)
        self.__key: Optional[int] = None
        self.__sdram_partition: Optional[
            SourceSegmentedSDRAMMachinePartition] = None
        self.__slice_index = slice_index
        self.__ring_buffer_shifts = ring_buffer_shifts
        self.__weight_scales = weight_scales
        self.__neuron_data = neuron_data
        self.__max_atoms_per_core = max_atoms_per_core
        self.__regenerate_data = False

    @property
    def _vertex_slice(self) -> Slice:
        return self.vertex_slice

    @property
    @overrides(PopulationMachineNeurons._slice_index)
    def _slice_index(self) -> int:
        return self.__slice_index

    @property
    @overrides(PopulationMachineNeurons._key)
    def _key(self) -> int:
        if self.__key is None:
            raise RuntimeError("key not yet set")
        return self.__key

    @property
    @overrides(PopulationMachineNeurons._has_key)
    def _has_key(self) -> bool:
        return self.__key is not None

    @overrides(PopulationMachineNeurons._set_key)
    def _set_key(self, key: int) -> None:
        self.__key = key

    @property
    @overrides(PopulationMachineNeurons._neuron_regions)
    def _neuron_regions(self) -> NeuronRegions:
        return self.NEURON_REGIONS

    @property
    @overrides(PopulationMachineNeurons._neuron_data)
    def _neuron_data(self) -> NeuronData:
        return self.__neuron_data

    @property
    @overrides(PopulationMachineNeurons._max_atoms_per_core)
    def _max_atoms_per_core(self) -> int:
        return self.__max_atoms_per_core

    def set_sdram_partition(
            self,
            sdram_partition: SourceSegmentedSDRAMMachinePartition) -> None:
        """ Sets the SDRAM Partition to receive data from.

        :param sdram_partition: The partition to receive data from.
        """
        if self.__sdram_partition is not None:
            raise SynapticConfigurationException(
                "Trying to set SDRAM partition more than once")
        self.__sdram_partition = sdram_partition

    @overrides(PopulationMachineCommon.parse_extra_provenance_items)
    def parse_extra_provenance_items(
            self, label: str, x: int, y: int, p: int,
            provenance_data: Sequence[int]) -> None:
        self._parse_neuron_provenance(
            x, y, p, provenance_data[:NeuronProvenance.N_ITEMS])

        neuron_prov = NeuronMainProvenance(
            *provenance_data[-NeuronMainProvenance.N_ITEMS:])

        with ProvenanceWriter() as db:
            db.insert_core(x, y, p, "Timer tick overruns",
                           neuron_prov.n_timer_overruns)
            if neuron_prov.n_timer_overruns > 0:
                db.insert_report(
                    f"Vertex {label} overran on "
                    f"{neuron_prov.n_timer_overruns} timesteps. "
                    f"This may mean that the simulation results are invalid."
                    " Try with fewer neurons per core, increasing the time"
                    " scale factor, or reducing the number of spikes sent")

    @overrides(PopulationMachineCommon.get_recorded_region_ids)
    def get_recorded_region_ids(self) -> List[int]:
        ids = self._pop_vertex.neuron_recorder.recorded_ids_by_slice(
            self.vertex_slice)
        return ids

    @overrides(AbstractGeneratesDataSpecification.generate_data_specification)
    def generate_data_specification(self, spec: DataSpecificationGenerator,
                                    placement: Placement) -> None:
        assert self.__sdram_partition is not None
        rec_regions = self._pop_vertex.neuron_recorder.get_region_sizes(
            self.vertex_slice)
        self._write_common_data_spec(spec, rec_regions)

        self._write_neuron_data_spec(spec, self.__ring_buffer_shifts)

        # Write information about SDRAM
        spec.reserve_memory_region(
            region=self.REGIONS.SDRAM_EDGE_PARAMS,
            size=SDRAM_PARAMS_SIZE, label="SDRAM Params")
        spec.switch_write_focus(self.REGIONS.SDRAM_EDGE_PARAMS)
        spec.write_value(
            self.__sdram_partition.get_sdram_base_address_for(self))
        spec.write_value(self.n_bytes_for_transfer)
        spec.write_value(len(self.__sdram_partition.pre_vertices))

        # End the writing of this specification:
        spec.end_specification()

    @overrides(
        AbstractRewritesDataSpecification.regenerate_data_specification)
    def regenerate_data_specification(self, spec: DataSpecificationReloader,
                                      placement: Placement) -> None:
        # Write the other parameters
        self._rewrite_neuron_data_spec(spec)

        # close spec
        spec.end_specification()

    @overrides(AbstractRewritesDataSpecification.reload_required)
    def reload_required(self) -> bool:
        return self.__regenerate_data

    @overrides(AbstractRewritesDataSpecification.set_reload_required)
    def set_reload_required(self, new_value: bool) -> None:
        self.__regenerate_data = new_value

    @property
    @overrides(ReceivesSynapticInputsOverSDRAM.weight_scales)
    def weight_scales(self) -> NDArray[floating]:
        return self.__weight_scales

    @staticmethod
    def get_n_bytes_for_transfer(n_neurons: int, n_synapse_types: int) -> int:
        """
        :param n_neurons: number of neurons
        :param n_synapse_types: number of synapse types
        :returns: The number of bytes needed for a transfer.
        """
        n_bytes = (2 ** get_n_bits(n_neurons) *
                   n_synapse_types *
                   ReceivesSynapticInputsOverSDRAM.N_BYTES_PER_INPUT)
        # May need to add some padding if not a round number of words
        extra_bytes = n_bytes % BYTES_PER_WORD
        if extra_bytes:
            n_bytes += BYTES_PER_WORD - extra_bytes
        return n_bytes

    @property
    @overrides(ReceivesSynapticInputsOverSDRAM.n_bytes_for_transfer)
    def n_bytes_for_transfer(self) -> int:
        return self.get_n_bytes_for_transfer(
            self.__max_atoms_per_core,
            self._pop_vertex.neuron_impl.get_n_synapse_types())

    @overrides(ReceivesSynapticInputsOverSDRAM.sdram_requirement)
    def sdram_requirement(self, sdram_machine_edge: SDRAMMachineEdge) -> int:
        if isinstance(sdram_machine_edge.pre_vertex,
                      SendsSynapticInputsOverSDRAM):
            return self.n_bytes_for_transfer
        raise SynapticConfigurationException(
            f"Unknown pre vertex type in edge {sdram_machine_edge}")

    @overrides(PopulationMachineNeurons.set_do_neuron_regeneration)
    def set_do_neuron_regeneration(self) -> None:
        self.__regenerate_data = True
        self.__neuron_data.reset_generation()

    @overrides(MachineVertex.get_n_keys_for_partition)
    def get_n_keys_for_partition(self, partition_id: str) -> int:
        n_colours = 2 ** self._pop_vertex.n_colour_bits
        return self.vertex_slice.n_atoms * n_colours



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from typing import List, Tuple, cast
from spinn_utilities.config_holder import get_config_bool
from spinn_utilities.log import FormatAdapter
from spinn_utilities.progress_bar import ProgressBar
from spinnman.model.enums import ExecutableType, CPUState, UserRegister
from spinnman.model import ExecutableTargets
from pacman.model.placements import Placement
from spinn_front_end_common.utilities.system_control_logic import (
    run_system_application)
from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.models.abstract_models import (
    AbstractSynapseExpandable, SYNAPSE_EXPANDER_APLX)

logger = FormatAdapter(logging.getLogger(__name__))


def synapse_expander() -> None:
    """
    Run the synapse expander.

    .. note::
        Needs to be done after data has been loaded.
    """
    # Find the places where the synapse expander and delay receivers should run
    expander_cores, expanded_placements, timeout = _plan_expansion()

    if expander_cores.total_processors:
        with ProgressBar(expander_cores.total_processors,
                         "Expanding Synapses") as progress:
            expander_app_id = SpynnakerDataView.get_new_id()
            run_system_application(
                expander_cores, expander_app_id,
                get_config_bool("Reports", "write_expander_iobuf") or False,
                None, frozenset({CPUState.FINISHED}), False,
                "synapse_expander_on_{}_{}_{}.txt",
                progress_bar=progress, logger=logger, timeout=timeout)

    # Once expander has run, fill in the connection data.
    if expanded_placements:
        with ProgressBar(len(expanded_placements),
                         "Reading generated connections") as progress:
            for placement in progress.over(expanded_placements):
                vertex = cast(AbstractSynapseExpandable, placement.vertex)
                vertex.read_generated_connection_holders(placement)


def _plan_expansion() -> Tuple[ExecutableTargets, List[Placement], float]:
    """
    Plan the expansion of synapses and set up the regions using USER1.

    :return: The places to load the synapse expander and delay expander
        executables, the target machine vertices to read synapses back from,
        and an estimated timeout for how long the expansion should be let run.
    """
    synapse_bin = SpynnakerDataView.get_executable_path(SYNAPSE_EXPANDER_APLX)
    expander_cores = ExecutableTargets()
    expanded_placements = list()
    txrx = SpynnakerDataView.get_transceiver()

    max_data = 0
    max_bit_field = 0
    to_write: List[Tuple[int, int, int, UserRegister, int]] = []
    for placement in SpynnakerDataView.iterate_placemements():
        # Add all machine vertices of the population vertex to ones
        # that need synapse expansion
        vertex = placement.vertex
        if isinstance(vertex, AbstractSynapseExpandable):
            if vertex.gen_on_machine():
                expander_cores.add_processor(
                    synapse_bin, placement.x, placement.y, placement.p,
                    executable_type=ExecutableType.SYSTEM)
                expanded_placements.append(placement)
                # Write the region to USER1, as that is the best we can do
                to_write.append((
                    *placement.location, UserRegister.USER_1,
                    vertex.connection_generator_region))
                max_data = max(max_data, vertex.max_gen_data)
                max_bit_field = max(max_bit_field, vertex.bit_field_size)

    txrx.write_user_many(to_write, "Preparing to Expand Synapses")

    # Allow 1 seconds per ~1000 synapses, with minimum of 2 seconds
    timeout = max(2.0, max_data / 1000.0)
    # Also allow 1s per 1000 bytes of bitfields
    timeout += max(2.0, max_bit_field / 1000.0)
    return expander_cores, expanded_placements, timeout
